//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// The code below is authored by Rafael Grompone, see code "Line 
// Segment Detector" of IPOL. See also README file. 

// The code has been adapted to the command-line parameters needed
// in this application. Adapted by Luis Garrido, <lluis.garrido@ub.edu>

#ifndef CMDLINE_H
#define CMDLINE_H

/*----------------------------------------------------------------------------*/
/** Definition of the command line interface.                                 */

#define USE "                                                                    \
#name: mgtn                                                                      \
#author: Lluis Garrido and Mostafa Kalmoun                                       \
#version: 17, December 2014                                                      \
#year: 2008-2014                                                                 \
#desc: MultiGrid Truncated Newton Optical Flow                                   \
#opt: model | m | int | 2 | 1 | 4 |                                              \
      Model 1, 2, 3 or 4 as described in the paper.                              \
#opt: scheme | s | int | 2 | 1 | 2 |                                             \
      Numerical scheme (1 = Multiresolution, 2 = Full Multigrid).                \
#opt: alpha | a | double | 50.0 | 10.0 | 10000.0 |                               \
      Weight for the regularizer.                                                \
#opt: vcycles | v | int | 5 | 1 | 25 |                                           \
      Number of V-cycles to perform at each level (only multigrid)               \
#opt: conffile | c | str | | | |                                                 \
      Configuration file (overrides previous command-line parameters).           \
#opt: gtruth | t | str | | | |                                                   \
      File containing ground truth in flo format.                                \
#opt: backcomp | b | str | | | |                                                 \
      Output backwards compensated image in file backcomp in PGM format.         \
#opt: quiet | q | int | 0 | 0 | 1 |                                              \
      Set to 1 to suppress messages to screen.                                   \
#req: inT0 | | str | | | | Input image at T = 0, PGM file expected.              \
#req: inT1 | | str | | | | Input image at T = 1, PGM file expected.              \
#req: outF | | str | | | | Output flow in flo format."
/*----------------------------------------------------------------------------*/

#define FIELD_LENGTH 160
/*----------------------------------------------------------------------------*/
/** Structure to store one argument definition and read value.
 */
struct argument
{
  char name[FIELD_LENGTH];     /* name to internally identify the argument */
  char desc[FIELD_LENGTH];     /* description */
  char id;                     /* letter used with '-' to use the option */
  char type;                   /* i=int, d=double, s=str, b=bool */
  int required;
  int assigned;
  int def_value;               /* true or false, a default value is assigned? */
  char d_value[FIELD_LENGTH];  /* default value */
  char s_value[FIELD_LENGTH];  /* string found, also the value if 'str' */
  int  i_value;
  double f_value;
  int    min_set;              /* true or false, is minimal value set? */
  double min;
  int    max_set;              /* true or false, is maximal value set? */
  double max;
};


/*----------------------------------------------------------------------------*/
/** Structure to store the full set of argument definitions and its values.
 */
struct arguments
{
  char name[FIELD_LENGTH];
  char author[FIELD_LENGTH];
  char version[FIELD_LENGTH];
  char desc[FIELD_LENGTH];
  char compiled[FIELD_LENGTH];
  char year[FIELD_LENGTH];
  int  arg_num;
  int  arg_allocated;
  struct argument * args;
};

/* Prototypes */
void free_arguments(struct arguments *arg);                                                                                                                                                   
struct arguments *process_arguments(char *desc, int argc, char **argv);                                                                                                                       
char *get_str(struct arguments *arg, char *name);                                                                                                                                             
int get_int(struct arguments *arg, char *name);                                                                                                                                               
double get_double(struct arguments *arg, char *name);
 

#endif
