//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include <stdio.h>
#include <stdlib.h>

#include "mgtn.h"

/**
 *
 *  Dense motion field 
 *
 */

void dense_motion_field(
  struct resources *DD)
{
  /* Prefilter images. This step is currently done only for the case the linear
   * OFE is used.
   */

  printMsg("Applying prefilter to original images...\n");
  if (DD->param.gradient_filter_type == FILTER_OFE_HS)
     mgtn_image_prefilter_images(DD->ori_in0, DD->ori_in1, DD);

  /* Initialize gradient images. The gradient computation includes
   * prefiltering in the filter taps (but original images are really
   * not prefiltered). 
   */

  mgtn_gradient_images_initialize(DD);

  /* Prefilter input images. This step has currently to be performed after the
   * image gradient computation, since the gradient computation uses filter
   * taps that includes prefiltering. In case linear OFE is used the
   * prefiltering is performed previous to construt the pyramid and previous to
   * compute the gradient. */

  if ((DD->param.gradient_filter_type != FILTER_OFE_HS) && (!DD->in_multigrid))
  {
     printMsg("Applying prefilter to original images...\n");
     mgtn_image_prefilter_images(DD->ori_in0, DD->ori_in1, DD);
  }

  /* Compute image pyramid, input images have been prefiltered. */

  mgtn_image_compute_pyramid(DD);

  /* Prefilter input images for the case of multigrid. In this case it is
   * numerically more stable to filter each image separately!
   */

  if ((DD->param.gradient_filter_type != FILTER_OFE_HS) && (DD->in_multigrid))
  {
    int i;
    printMsg("Applying prefilter to original images...\n");
    for(i = DD->param.max_level; i >= 0; i--)
      mgtn_image_prefilter_images(DD->in0[i], DD->in1[i], DD);
  }

  /* Motion estimation procedure begins here! */

  clock_t t1, t2;
  float elapsed_time, total_elapsed_time;

  double f;
  int i, rc, uv_initialized, nb_iterations;

  uv_initialized = 0;
  total_elapsed_time = 0.0;

  /* Initialization of motion estimation procedure */

  switch (DD->param.scheme)
  {
    case SCHEME_MULTIRESOLUTION:  /* Multiresolution (starting at coarsest level) */

      DD->in_multigrid    = 0;
      DD->in_presmoothing = 0;
      break;

    case SCHEME_FMG:             /* Full multigrid (starting at coarsest level) */

      DD->in_multigrid    = 1;
      DD->in_presmoothing = 0;

      if (DD->param.type_multigrid == V_CYCLE)
	mgtn_initialize_v_cycle(DD);
      else // W_CYCLE
	mgtn_initialize_w_cycle(DD);

      break;
  }

  nb_iterations = 1;   /* Avoid complaining of compiler */

  /* This is the loop that runs through all resolution levels */

  for(i = DD->param.max_level; i >= 0; i--)
  {
    /* curr_level is a variable used to know at which level we
       are currently computing optical flow */

    DD->curr_level = i;

    /* Initialize level in case we have selected a multigrid */

    if (DD->in_multigrid)
    {

      if (DD->param.type_multigrid == V_CYCLE)
	mgtn_generate_v_cycle(DD);
      else // W_CYCLE
	mgtn_generate_w_cycle(DD);

      /* nb_iterations is the number of V or W-cycles that
	 has to be performed. */
      nb_iterations = DD->param.maxit_v_cycle[i];

      if ((!nb_iterations) && (!uv_initialized))
	continue;
    }

    /* Initialize resources for the presmooth in case we are using
       multigrid. The resources are the same as the needed for a simple
       multiresolution. */

    mgtn_resources_init_presmooth(DD);

    /* The first time the condition is executed the flow will be initialized.
       Otherwise the flow will be prolongated from the current level to the
       next finer level. */  

    if (!uv_initialized)
    {
      mgtn_flow_initialize(DD);
      mgtn_set_border_conditions_uv(GET_CURRENT_IMAGE(DD,u), GET_CURRENT_IMAGE(DD,v));
      uv_initialized = 1;
    }
    else
    {
      /* The flow is prolongated from the coarser level to the current level */ 

      mgtn_image_prolongation2(GET_COARSER_IMAGE(DD, u), GET_COARSER_IMAGE(DD, v), GET_CURRENT_IMAGE(DD, u), GET_CURRENT_IMAGE(DD, v), DD);
      mgtn_set_border_conditions_uv(GET_CURRENT_IMAGE(DD,u), GET_CURRENT_IMAGE(DD,v));

      /* We now delete the resources of the coarser level */

      DD->curr_level++;
      mgtn_resources_delete_presmooth(DD);
      mgtn_resources_delete_postsmooth(DD); 
      DD->curr_level--;
    }

    if (!nb_iterations)
      continue;

    /* Everything has been initialized for the current level of resolution.
       Now perform optical flow.*/

    switch (DD->param.scheme)
    {
      case SCHEME_MULTIRESOLUTION:
	printMsg("In level %d...", i);
	fflush(stdout);

	t1 = clock();

	f = mgtn_functional_evaluation(GET_CURRENT_IMAGE(DD, u), GET_CURRENT_IMAGE(DD, v), DD);
	mgtn_gradient_evaluation(GET_CURRENT_IMAGE(DD, u), GET_CURRENT_IMAGE(DD, v), GET_CURRENT_IMAGE(DD, grad_E_u), GET_CURRENT_IMAGE(DD, grad_E_v), DD);

	DD->nfeval[i]++;
	DD->ngeval[i]++;

	/* Truncated Newton Method */

	mgtn_tnc(&f, GET_CURRENT_IMAGE(DD, u), GET_CURRENT_IMAGE(DD, v), GET_CURRENT_IMAGE(DD, grad_E_u), GET_CURRENT_IMAGE(DD, grad_E_v),
	    DD->param.maxit_multiresolution, 0, DD);

	t2 = clock();

	elapsed_time = (float) ((t2 - t1) / (float) CLOCKS_PER_SEC);
	total_elapsed_time += elapsed_time;

	/* Test if finished */

	printMsg("\nElapsed time in this level: %f sec\n", elapsed_time);

	break;

      case SCHEME_FMG:

	/* The V or W-cycle is done nb_iterations times */

	for(int j = 0; j < nb_iterations; j++)
	{
	  printMsg("Cycle (%d of %d) at level %d...", j+1, nb_iterations, i);
	  fflush(stdout);

	  /* Initialize clocks. We will not count loading data (i.e. the
	   * gradients of the images) from disc. The elapsed time will be
	   * stored in t_elapsed */

	  DD->t_elapsed = 0;
	  DD->t_active = 1;
	  DD->t_last = t1 = clock();

	  /* This function performs the V or W-cycle. The return value rc allows
	     to know if we have converged. */

	  rc = mgtn_multigrid(DD);

	  t2 = clock();
	  DD->t_active = 0;

	  elapsed_time  = (float) (DD->t_elapsed + (t2 - DD->t_last)) / ((float) CLOCKS_PER_SEC);
	  total_elapsed_time += elapsed_time;

	  printMsg("\nElapsed time in this cycle: %f sec\n", elapsed_time);

	  /* Test if converged for the current level */

	  if (rc == MGTN_TNC_CONVERGED)
	    j = nb_iterations; 
	}
    }
  }

  printMsg("Total elapsed time: %f sec\n", total_elapsed_time);
  printMsg("\nFunction and gradient evaluations:\n");

  double kappa;
  double nfunc = 0.0;
  double ngrad = 0.0;

  for(i = 0; i <= DD->param.max_level; i++)
  {
    kappa = pow(2.0, - 2.0 * (double) i);
    nfunc += (double) DD->nfeval[i] * kappa;
    ngrad += (double) DD->ngeval[i] * kappa;
    printMsg("Level %d, function evaluations: %d, gradient evaluations: %d\n", i, DD->nfeval[i], DD->ngeval[i]);
  }

  printMsg("Accumulated function evaluations along the levels: %f\n", (float) nfunc);
  printMsg("Accumulated gradient evaluations along the levels: %f\n", (float) ngrad);
}



