//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 *
 *  This function is called by the TNC (Truncated Newton) library to evaluate
 *  the energy in case we are using multigrid. 
 *
 */

double mgtn_functional_evaluation_multigrid(
  struct image *u, 
  struct image *v,
  struct image *residue_u, 
  struct image *residue_v,
  struct resources *DD)
{
  double cost, cost_v;

  cost   = mgtn_functional_evaluation(u, v, DD);
  cost_v = mgtn_functional_evaluation_residue(u, v, residue_u, residue_v);

  cost -= cost_v;

  return cost;
}

/**
 *
 *  This function is used internally by mgtn_functional_evaluation_v_cycle. Do not
 *  call directly. 
 *
 */

double mgtn_functional_evaluation_residue(
  struct image *u,
  struct image *v,
  struct image *residue_u, 
  struct image *residue_v)
{
  int i, j, ncol, nrow, offset;
  int min_x, min_y, max_x, max_y;

  double cost_v, *p_base1, *p_base2, *p_residue, *p_x;

  cost_v = 0.0;

  ncol = u->ncol;
  nrow = u->nrow;

  min_x = 1;
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  p_base1 = residue_u->gray;
  p_base2 = u->gray;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;

    p_residue = p_base1 + offset;
    p_x       = p_base2 + offset;

    for(j = min_x; j <= max_x; j++, p_residue++, p_x++)
       cost_v += *p_residue * *p_x;
  }

  p_base1 = residue_v->gray;
  p_base2 = v->gray;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;

    p_residue = p_base1 + offset;
    p_x       = p_base2 + offset;

    for(j = min_x; j <= max_x; j++, p_residue++, p_x++)
       cost_v += *p_residue * *p_x;
  }

  return cost_v;
}

/**
 *
 *  This is the function that evaluates the energy of the function. It is
 *  called by TNC in case we are using a multiresolution or via
 *  mgtn_functional_evaluation_multigrid in case we are using multigrid.
 *
 */

double mgtn_functional_evaluation(
  struct image *u,
  struct image *v, 
  struct resources *DD)
{
  struct vector2 coord_phi, coord_pix;

  int res, nrow, ncol;
  int max_x, max_y, min_x, min_y;

  int offset, offset_tmp;

  double cost_data, cost_regul;
  double cost, term1, term2, term3, term4;
  double alpha, motder[8]; 

  double *p_u, *p_v;

  /* Initialize nrow, ncol and other stuff */

  nrow = GET_CURRENT_NROW(DD);
  ncol = GET_CURRENT_NCOL(DD);

  min_x = 1;
  min_y = 1;

  max_x = ncol - 1;
  max_y = nrow - 1;

  /* Compute the "distance" between pixels for the current
   * resolution level. */

  res = DD->curr_level;
  DD->inverse_h = pow(0.5, (double) res);

  /* Initalize cost */

  cost = cost_data = cost_regul = 0;
  
  coord_phi.x = 0.0; // Avoid complaining of compiler
  coord_phi.y = 0.0;

  /* Evaluate data term. Energy is evaluated only in the interior of the
   * image, i.e. it is not evaluated on the outer border of the image, since
   * this border is used for the border conditions. 
   */

  for(coord_pix.y = min_y; coord_pix.y < max_y; coord_pix.y++)
  {
    offset = coord_pix.y * ncol + min_x;

    p_u = u->gray + offset;
    p_v = v->gray + offset;

    for(coord_pix.x = min_x; coord_pix.x < max_x; coord_pix.x++, p_u++, p_v++)
    {
	switch (DD->param.functional_data_type)
	{
	  case FUNCTIONAL_DATA_NON_LINEAR:

	    coord_phi.x = coord_pix.x + *p_u * DD->inverse_h;
	    coord_phi.y = coord_pix.y + *p_v * DD->inverse_h;

	    if ((coord_phi.x >= 0) && (coord_phi.x <= max_x) && 
		(coord_phi.y >= 0) && (coord_phi.y <= max_y))
	    {
	      term1 = mgtn_image_interpolate(0, &coord_pix, DD);
	      term2 = mgtn_image_interpolate(1, &coord_phi, DD);
	      term3 = term2 - term1;

	      if (fabs(term3) > DD->param.outlier)
		term3 = DD->param.outlier;

	      cost_data += 0.5 * SQR(term3);
	    }

	    break;

	  case FUNCTIONAL_DATA_LINEAR:

	    offset_tmp = coord_pix.x + coord_pix.y * ncol;

	    term1 = DD->grad_t0_x[res]->gray[offset_tmp];
	    term2 = DD->grad_t0_y[res]->gray[offset_tmp];
	    term3 = DD->grad_t1_x[res]->gray[offset_tmp];

	    term4 = term1 * *p_u + term2 * *p_v + term3; 

	    if (fabs(term4) > DD->param.outlier) 
	      term4 = DD->param.outlier;

	    cost_data += 0.5 * SQR(term4);

	    break;
	}

	/* Get motion derivatives. Motion derivatives are computed using
	 * forward/backward differences. */

	mgtn_get_motion_derivatives(p_u, p_v, motder, DD);

	/* Compute regularization term. In our implementation the border pixels
	 * are used to store the border conditions (i.e. the border values of u
	 * and v). It is implicitly assumed that Neumann boundary conditions are
	 * satisfied.
	 */

	term1 = 
	  0.5 * SQR(motder[VXf]) + 0.5 * SQR(motder[VXb]) +
	  0.5 * SQR(motder[VYf]) + 0.5 * SQR(motder[VYb]) +
	  0.5 * SQR(motder[UXf]) + 0.5 * SQR(motder[UXb]) +
	  0.5 * SQR(motder[UYf]) + 0.5 * SQR(motder[UYb]);

	switch (DD->param.functional_regul_type)
	{
	  case FUNCTIONAL_REGUL_QUADRATIC:  /* Horn - Schunk regularization */

	    cost_regul += term1;
	    break;

	  case FUNCTIONAL_REGUL_TV:  /* Total variation regularization */

	    cost_regul += sqrt(term1 + SQR(DD->param.mu));
	    break;
	}
    }
  }

  alpha = DD->param.alpha;
  cost  = cost_data + alpha * cost_regul;

  return cost;
}

