//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 *
 *  This function is called by the TNC (Truncated Newton) library to evaluate
 *  the energy in case we are using multigrid. 
 *
 */

void mgtn_gradient_evaluation_multigrid(
  struct image *u, 
  struct image *v, 
  struct image *residue_u, 
  struct image *residue_v, 
  struct image *grad_E_u, 
  struct image *grad_E_v,
  struct resources *DD)
{
  mgtn_gradient_evaluation(u, v, grad_E_u, grad_E_v, DD);
  mgtn_gradient_evaluation_residue(grad_E_u, grad_E_v, residue_u, residue_v);
}

/**
 *
 *  This function is called internally by mgtn_gradient_evaluation_multigrid. Do not
 *  call directly.
 *
 */

void mgtn_gradient_evaluation_residue(
  struct image *grad_E_u,
  struct image *grad_E_v, 
  struct image *residue_u, 
  struct image *residue_v)
{
  int i, j, ncol, nrow, offset;
  int min_x, min_y, max_x, max_y;

  double *p_base1, *p_base2, *p_grad, *p_residue;

  ncol = grad_E_u->ncol;
  nrow = grad_E_u->nrow;

  min_x = 1; 
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  p_base1 = grad_E_u->gray;
  p_base2 = residue_u->gray;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;

    p_grad    = p_base1 + offset;
    p_residue = p_base2 + offset;

    for(j = min_x; j <= max_x; j++, p_grad++, p_residue++)
      *p_grad -= *p_residue;
  }

  p_base1 = grad_E_v->gray;
  p_base2 = residue_v->gray;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;

    p_grad    = p_base1 + offset;
    p_residue = p_base2 + offset;

    for(j = min_x; j <= max_x; j++, p_grad++, p_residue++)
      *p_grad -= *p_residue;
  }
}

/**
 *
 *  This function evaluates the gradient of the function. It is called by TNC if
 *  multiresolution is used or via mgtn_gradient_evaluation_multigrid in case we
 *  are in multigrid. 
 *
 */

void mgtn_gradient_evaluation(
  struct image *u, 
  struct image *v, 
  struct image *grad_E_u, 
  struct image *grad_E_v, 
  struct resources *DD)
{
  struct vector2 coord_pix;

  int ncol, nrow, offset;
  int min_x, min_y, max_x, max_y;

  double alpha;
  double *p_grad_E_u, *p_grad_E_v;
  double grad_data_u, grad_data_v, grad_reg_u, grad_reg_v;

  /* Initialization */

  nrow  = u->nrow;
  ncol  = u->ncol;

  min_x = 1;
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  /* Clear images with zero values */

  mgtn_image_clear(grad_E_u, 0.0);
  mgtn_image_clear(grad_E_v, 0.0);

  /* Initialize variables */

  DD->inverse_h = pow(0.5, (double) DD->curr_level);

  alpha = DD->param.alpha;

  for(coord_pix.y = min_y; coord_pix.y <= max_y; coord_pix.y++)
  {
    offset = coord_pix.y * ncol + min_x;

    p_grad_E_u = grad_E_u->gray + offset;
    p_grad_E_v = grad_E_v->gray + offset;

    for(coord_pix.x = min_x; coord_pix.x <= max_x; coord_pix.x++, p_grad_E_u++, 
	p_grad_E_v++)
    {
      /* Avoid complaining of compiler */

      grad_data_u = grad_data_v = 0.0;
      grad_reg_u  = grad_reg_v  = 0.0;

      /* Call gradient evaluations for the data and regularization terms */

      mgtn_gradient_evaluation_data(&coord_pix, u, v, &grad_data_u, &grad_data_v, DD);
      mgtn_gradient_evaluation_reg(&coord_pix, u, v, &grad_reg_u, &grad_reg_v, DD);

      /* Store result */

      *p_grad_E_u = grad_data_u + alpha * grad_reg_u;
      *p_grad_E_v = grad_data_v + alpha * grad_reg_v;
    }
  }
}

/**
 *
 * This function is called by mgtn_gradient_evaluation. Do
 * not call directly. 
 *
 */

void mgtn_gradient_evaluation_data(
  struct vector2 *coord_pix, 
  struct image *u, 
  struct image *v, 
  double *grad_data_u, 
  double *grad_data_v,
  struct resources *DD)
{
  struct vector2 coord_phi, grad_f_t1;

  double term1, term2, term3, term4;

  double *p_u, *p_v;

  int res, nrow, ncol, max_x, max_y, offset;

  /* Initialize */

  *grad_data_u = 0.0;
  *grad_data_v = 0.0;

  res = DD->curr_level;

  nrow = u->nrow;
  ncol = u->ncol;

  max_x = ncol - 1;
  max_y = nrow - 1;

  switch (DD->param.functional_data_type)
  {
    case FUNCTIONAL_DATA_NON_LINEAR:

      offset = coord_pix->y * ncol + coord_pix->x;

      p_u = u->gray + offset;
      p_v = v->gray + offset;

      coord_phi.x = coord_pix->x + *p_u * DD->inverse_h;
      coord_phi.y = coord_pix->y + *p_v * DD->inverse_h;

      /* point may be out of the image domain. */

      if ((coord_phi.x < 0) || (coord_phi.x > max_x) ||
	  (coord_phi.y < 0) || (coord_phi.y > max_y))
	return;

      term1 = mgtn_image_interpolate(0, coord_pix, DD);
      term2 = mgtn_image_interpolate(1, &coord_phi, DD);
      term3 = term2 - term1;

      mgtn_gradient_images_interpolate_first(1, &coord_phi, &grad_f_t1, DD);

      if (fabs(term3) < DD->param.outlier)
      {
	*grad_data_u += grad_f_t1.x * term3;
	*grad_data_v += grad_f_t1.y * term3;
      }

      break;

    case FUNCTIONAL_DATA_LINEAR:

      offset = (int) coord_pix->x + (int) coord_pix->y * ncol;

      p_u = u->gray + offset;
      p_v = v->gray + offset;

      term1 = DD->grad_t0_x[res]->gray[offset];
      term2 = DD->grad_t0_y[res]->gray[offset];
      term3 = DD->grad_t1_x[res]->gray[offset];

      term4 = term1 * *p_u + term2 * *p_v + term3; 

      if (fabs(term4) < DD->param.outlier) 
      {
	*grad_data_u += term4 * term1;
	*grad_data_v += term4 * term2;
      }

      break;
  }
}

/**
 *
 * This function is called by mgtn_gradient_evaluation. Do
 * not call directly. 
 *
 */

void mgtn_gradient_evaluation_reg(
  struct vector2 *coord_pix, 
  struct image *u, 
  struct image *v, 
  double *grad_reg_u, 
  double *grad_reg_v,
  struct resources *DD)
{
  /* HS */
  double a0, a1, a2, a3, a4;
  /* TV */
  double mu2;
  /* TV2 */
  double *p_u, *p_v;
  double motder[8];
  double term1, term2;

  double inverse_h;

  int nrow, ncol, max_x, max_y, offset;
  int m, n;

  *grad_reg_u = 0.0;
  *grad_reg_v = 0.0;

  nrow = u->nrow;
  ncol = u->ncol;

  max_x = ncol - 1;
  max_y = nrow - 1;

  m = coord_pix->x;
  n = coord_pix->y;
  offset = n * ncol + m;

  inverse_h = DD->inverse_h;

  switch (DD->param.functional_regul_type)
  {
    case FUNCTIONAL_REGUL_QUADRATIC:

      /* u component */

      a0 = u->gray[offset];
      a1 = u->gray[offset - ncol];
      a2 = u->gray[offset - 1];
      a3 = u->gray[offset + 1];
      a4 = u->gray[offset + ncol];

      *grad_reg_u = 2.0 * (4.0 * a0 - (a1 + a2 + a3 + a4));
      *grad_reg_u *= SQR(inverse_h);

      /* v component */

      a0 = v->gray[offset];
      a1 = v->gray[offset - ncol];
      a2 = v->gray[offset - 1];
      a3 = v->gray[offset + 1];
      a4 = v->gray[offset + ncol];

      *grad_reg_v = 2.0 * (4.0 * a0 - (a1 + a2 + a3 + a4));
      *grad_reg_v *= SQR(inverse_h);

      break;

    case FUNCTIONAL_REGUL_TV:

      mu2 = SQR(DD->param.mu);

      p_u = u->gray + offset;
      p_v = v->gray + offset;

      /* Go for (m,n) */

      mgtn_get_motion_derivatives(p_u, p_v, motder, DD);

      term1 = 
	0.5 * SQR(motder[VXf]) + 0.5 * SQR(motder[VXb]) +
	0.5 * SQR(motder[VYf]) + 0.5 * SQR(motder[VYb]) +
	0.5 * SQR(motder[UXf]) + 0.5 * SQR(motder[UXb]) +
	0.5 * SQR(motder[UYf]) + 0.5 * SQR(motder[UYb]);

      term2 = inverse_h / sqrt(term1 + mu2);

      *grad_reg_u = 0.5 * (motder[UXb] + motder[UYb] - motder[UXf] - motder[UYf]) * term2;
      *grad_reg_v = 0.5 * (motder[VXb] + motder[VYb] - motder[VXf] - motder[VYf]) * term2;

      /* Go for (m-1,n) */

      p_u--;
      p_v--;
      m--;

      if (m != 0) {
	mgtn_get_motion_derivatives(p_u, p_v, motder, DD);

	term1 = 
	  0.5 * SQR(motder[VXf]) + 0.5 * SQR(motder[VXb]) +
	  0.5 * SQR(motder[VYf]) + 0.5 * SQR(motder[VYb]) +
	  0.5 * SQR(motder[UXf]) + 0.5 * SQR(motder[UXb]) +
	  0.5 * SQR(motder[UYf]) + 0.5 * SQR(motder[UYb]);

	term2 = inverse_h / sqrt(term1 + mu2);

	*grad_reg_u += 0.5 * motder[UXf] * term2;
	*grad_reg_v += 0.5 * motder[VXf] * term2;
      }

      p_u++;
      p_v++;
      m++;

      /* Go for (m,n-1) */

      p_u -= ncol;
      p_v -= ncol;
      n--;

      if (n != 0) {
	mgtn_get_motion_derivatives(p_u, p_v, motder, DD);

	term1 = 
	  0.5 * SQR(motder[VXf]) + 0.5 * SQR(motder[VXb]) +
	  0.5 * SQR(motder[VYf]) + 0.5 * SQR(motder[VYb]) +
	  0.5 * SQR(motder[UXf]) + 0.5 * SQR(motder[UXb]) +
	  0.5 * SQR(motder[UYf]) + 0.5 * SQR(motder[UYb]);

	term2 = inverse_h / sqrt(term1 + mu2);

	*grad_reg_u += 0.5 * motder[UYf] * term2;
	*grad_reg_v += 0.5 * motder[VYf] * term2;
      }

      p_u += ncol;
      p_v += ncol;
      n++;

      /* Go for (m+1,n) */

      p_u++;
      p_v++;
      m++;

      if (m != max_x) {
	mgtn_get_motion_derivatives(p_u, p_v, motder, DD);

	term1 = 
	  0.5 * SQR(motder[VXf]) + 0.5 * SQR(motder[VXb]) +
	  0.5 * SQR(motder[VYf]) + 0.5 * SQR(motder[VYb]) +
	  0.5 * SQR(motder[UXf]) + 0.5 * SQR(motder[UXb]) +
	  0.5 * SQR(motder[UYf]) + 0.5 * SQR(motder[UYb]);

	term2 = inverse_h / sqrt(term1 + mu2);

	*grad_reg_u -= 0.5 * motder[UXb] * term2;
	*grad_reg_v -= 0.5 * motder[VXb] * term2;
      }

      p_u--;
      p_v--;
      m--;

      /* Go for (m,n+1) */

      p_u += ncol;
      p_v += ncol;
      n++;

      if (n != max_y) {
	mgtn_get_motion_derivatives(p_u, p_v, motder, DD);

	term1 = 
	  0.5 * SQR(motder[VXf]) + 0.5 * SQR(motder[VXb]) +
	  0.5 * SQR(motder[VYf]) + 0.5 * SQR(motder[VYb]) +
	  0.5 * SQR(motder[UXf]) + 0.5 * SQR(motder[UXb]) +
	  0.5 * SQR(motder[UYf]) + 0.5 * SQR(motder[UYb]);

	term2 = inverse_h / sqrt(term1 + mu2);

	*grad_reg_u -= 0.5 * motder[UYb] * term2;
	*grad_reg_v -= 0.5 * motder[VYb] * term2;
      }

      p_u -= ncol;
      p_v -= ncol;
      n--;

      break;
  }
}

