//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 *
 *  Function used to compute motion derivatives for a given point
 *
 */

int mgtn_get_motion_derivatives(
  double *p_u, 
  double *p_v, 
  double motder[8],
  struct resources *DD)
{
  int ncol;
  double inverse_h;

  ncol = GET_CURRENT_NCOL(DD);
  inverse_h = DD->inverse_h;

  motder[VXf]= (*(p_v + 1) - *p_v) * inverse_h;
  motder[VYf]= (*(p_v + ncol) - *p_v) * inverse_h;
  motder[UXf]= (*(p_u + 1) - *p_u) * inverse_h;
  motder[UYf]= (*(p_u + ncol) - *p_u) * inverse_h;

  motder[VXb]=(*p_v - *(p_v - 1)) * inverse_h;
  motder[VYb]=(*p_v - *(p_v - ncol)) * inverse_h;
  motder[UXb]=(*p_u - *(p_u - 1)) * inverse_h;
  motder[UYb]=(*p_u - *(p_u - ncol)) * inverse_h;

  return 0;
}

/**
 *
 *  This function is used to set the border conditions for the flow.  Indeed,
 *  in this implementation the border conditions are implemented through the
 *  pixels at the border of the image. Only Newman conditions are implemented 
 *  here.
 *
 */

void mgtn_set_border_conditions_uv(
  struct image *u, 
  struct image *v)
{
  int i, nrow, ncol, offset;
  int min_x, min_y, max_x, max_y;

  double *p_u, *p_v;

  nrow = u->nrow;
  ncol = v->ncol;

  min_x = 0;
  min_y = 0;

  max_x = ncol - 1;
  max_y = nrow - 1;

  offset = min_x + min_y * ncol;

  p_u = u->gray + offset; 
  p_v = v->gray + offset; 

  /* we are at the top left corner */

  i = min_x;

  *p_u = *(p_u + ncol + 1);
  *p_v = *(p_v + ncol + 1);

  for(i++, p_u++, p_v++; i < max_x; i++, p_u++, p_v++)
  {
    *p_u = *(p_u + ncol);
    *p_v = *(p_v + ncol);
  }

  /* we are at the top right corner */

  i = min_y;

  *p_u = *(p_u + ncol - 1);
  *p_v = *(p_v + ncol - 1);

  for(i++, p_u += ncol, p_v += ncol; i < max_y; i++, p_u += ncol, p_v += ncol)
  {
    *p_u = *(p_u - 1);
    *p_v = *(p_v - 1);
  }

  /* we are at the bottom right corner */

  i = max_x;

  *p_u = *(p_u - ncol - 1);
  *p_v = *(p_v - ncol - 1);

  for(i--, p_u--, p_v--; i > min_x; i--, p_u--, p_v--)
  {
    *p_u = *(p_u - ncol);
    *p_v = *(p_v - ncol);
  }

  /* we are the bottom left corner */

  i = max_y;

  *p_u = *(p_u - ncol + 1);
  *p_v = *(p_v - ncol + 1);

  for(i--, p_u -= ncol, p_v -= ncol; i > min_y; i--, p_u -= ncol, p_v -= ncol)
  {
    *p_u = *(p_u + 1);
    *p_v = *(p_v + 1);
  }

  /* done */
}

/**
 *
 *  This function is used to set the border conditions for the gradient. This
 *  function is only needed for multigrid.
 *
 */

void mgtn_set_border_conditions_gradient(
    struct image *grad_u, 
    struct image *grad_v) 
{
  int i, nrow, ncol, offset;
  int min_x, min_y, max_x, max_y;

  double *p_u, *p_v;

  nrow = grad_u->nrow;
  ncol = grad_v->ncol;

  min_x = 0;
  min_y = 0;

  max_x = ncol - 1;
  max_y = nrow - 1;

  offset = min_x + min_y * ncol;

  p_u = grad_u->gray + offset; 
  p_v = grad_v->gray + offset;

  /* we are at the top left corner */

  i = min_x;

  *p_u = *(p_u + ncol + 1);
  *p_v = *(p_v + ncol + 1);

  for(i++, p_u++, p_v++; i < max_x; i++, p_u++, p_v++)
  {
    *p_u = *(p_u + ncol);
    *p_v = *(p_v + ncol);
  }

  /* we are at the top right corner */

  i = min_y;

  *p_u = *(p_u + ncol - 1);
  *p_v = *(p_v + ncol - 1);

  for(i++, p_u += ncol, p_v += ncol; i < max_y; i++, p_u += ncol, p_v += ncol)
  {
    *p_u = *(p_u - 1);
    *p_v = *(p_v - 1);
  }

  /* we are at the bottom right corner */

  i = max_x;

  *p_u = *(p_u - ncol - 1);
  *p_v = *(p_v - ncol - 1);

  for(i--, p_u--, p_v--; i > min_x; i--, p_u--, p_v--)
  {
    *p_u = *(p_u - ncol);
    *p_v = *(p_v - ncol);
  }

  /* we are the bottom left corner */

  i = max_y;

  *p_u = *(p_u - ncol + 1);
  *p_v = *(p_v - ncol + 1);

  for(i--, p_u -= ncol, p_v -= ncol; i > min_y; i--, p_u -= ncol, p_v -= ncol)
  {
    *p_u = *(p_u + 1);
    *p_v = *(p_v + 1);
  }
}


