//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 *
 *  Computes the gradient images. The gradient at the finest level is computed
 *  by means of finite differences. The gradient the remaining levels is
 *  computed by means of restriction of the gradient with respect the previous
 *  level.  Images are saved to disc since they may be very large (this depends
 *  on the parameter D_INTERPOLATE defined in mgtn.h). Gradient images are
 *  loaded from disc each time they are needed.
 *
 */

void mgtn_gradient_images_initialize(
  struct resources *DD)
{
  int res, i;

  DD->curr_level = 0;
  printMsg("Computing gradients for level 0...\n");

  switch (DD->param.gradient_filter_type)
  {
    case FILTER_SIMONCELLI:
    case FILTER_GAUSS:
      mgtn_gradient_images_finite_differences(DD);
      break;

    case FILTER_OFE_HS:
      mgtn_gradient_images_hornschunk(DD);
      break;
  }

#ifdef LOW_MEMORY_COMPUTER
  mgtn_write_gradient_images(DD);
#endif

  for(i = 1; i <= DD->param.max_level; i++)
  {
    res = DD->curr_level = i;

    printMsg("Computing gradients for level %d using restriction...\n", res);
    mgtn_gradient_images_restriction(DD);

#ifdef LOW_MEMORY_COMPUTER
    mgtn_write_gradient_images(DD);

    DD->curr_level = res - 1;
    mgtn_gradient_images_delete(DD);
    DD->curr_level = res;
#endif
  }

#ifdef LOW_MEMORY_COMPUTER
  mgtn_gradient_images_delete(DD);
#endif
}

/**
 *
 * Frees memory allocated for gradient at current level 
 *
 */

void mgtn_gradient_images_delete(
    struct resources *DD)
{
  int res;

  res = DD->curr_level;

  if (DD->grad_t0_x[res])
  {
    mgtn_image_delete(DD->grad_t0_x[res]);
    mgtn_image_delete(DD->grad_t0_y[res]);
    mgtn_image_delete(DD->grad_t1_x[res]);
    mgtn_image_delete(DD->grad_t1_y[res]);

    DD->grad_t0_x[res] = NULL;
    DD->grad_t0_y[res] = NULL;
    DD->grad_t1_x[res] = NULL;
    DD->grad_t1_y[res] = NULL;

    if (DD->grad_t1_xx[res])
    {
      mgtn_image_delete(DD->grad_t1_xx[res]);
      mgtn_image_delete(DD->grad_t1_yy[res]);
      mgtn_image_delete(DD->grad_t1_xy[res]);

      DD->grad_t1_xx[res] = NULL;
      DD->grad_t1_yy[res] = NULL;
      DD->grad_t1_xy[res] = NULL;
    }
  }
}


/**
 *
 *  This function computes, for a given resolution level, the gradient of the
 *  image in the directions x and y. The function uses internally
 *  mgtn_gradient_images_finite_differences_convolution.
 *
 *  Note that the first and second derivatives are computed, although the
 *  second derivatives are really not needed for the current energy
 *  implementations. Thus we allow the user to implement other types of
 *  energies such as those based on the gradient-constancy assumption.
 *
 */

void mgtn_gradient_images_finite_differences(
    struct resources *DD)
{
  struct vector *grad0_x, *grad1_x, *grad0_y, *grad1_y;

  /* Compute gradient for images at t0. For t0 the gradient is not computed
   * with subpixel interpolation. Gradient is computed just at the integer
   * pixel position. In addition, only first order gradients are computed. 
   */

  /* Get filters for fist derivative */
  grad0_x = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_FIRST, DD->param.gradient_filter_size, 0, 1);
  grad1_x = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_FIRST, DD->param.gradient_filter_size, 1, 1);

  /* Compute gradients */
  DD->grad_t0_x[0] = mgtn_gradient_images_finite_differences_convolution(DD->ori_in0, grad1_x, grad0_x, 0, DD);
  DD->grad_t0_y[0] = mgtn_gradient_images_finite_differences_convolution(DD->ori_in0, grad0_x, grad1_x, 0, DD);

  /* Delete gradient filters */
  mgtn_vector_delete(grad0_x);
  mgtn_vector_delete(grad1_x);

  /* Compute gradient for images at t1. For t1 the gradient is computed with
   * subpixel interpolation. 
   */

  /* Get filters for fist derivative */
  grad0_x = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_FIRST, DD->param.gradient_filter_size, 0, DD->param.interpolated_samples);
  grad0_y = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_FIRST, DD->param.gradient_filter_size, 0, DD->param.interpolated_samples);
  grad1_x = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_FIRST, DD->param.gradient_filter_size, 1, DD->param.interpolated_samples);
  grad1_y = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_FIRST, DD->param.gradient_filter_size, 1, DD->param.interpolated_samples);

  /* Compute gradients */
  DD->grad_t1_x[0] = mgtn_gradient_images_finite_differences_convolution(DD->ori_in1, grad1_x, grad0_y, 1, DD);
  DD->grad_t1_y[0] = mgtn_gradient_images_finite_differences_convolution(DD->ori_in1, grad0_x, grad1_y, 1, DD);

  /* Delete */
  mgtn_vector_delete(grad0_x);
  mgtn_vector_delete(grad0_y);
  mgtn_vector_delete(grad1_x);
  mgtn_vector_delete(grad1_y);

  /* Get filters for second derivative */
  grad0_x = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_SECOND, DD->param.gradient_filter_size, 0, DD->param.interpolated_samples);
  grad0_y = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_SECOND, DD->param.gradient_filter_size, 0, DD->param.interpolated_samples);
  grad1_x = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_SECOND, DD->param.gradient_filter_size, 2, DD->param.interpolated_samples);
  grad1_y = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_SECOND, DD->param.gradient_filter_size, 2, DD->param.interpolated_samples);

  /* Compute gradients */
  DD->grad_t1_xx[0] = mgtn_gradient_images_finite_differences_convolution(DD->ori_in1, grad1_x, grad0_y, 1, DD);
  DD->grad_t1_yy[0] = mgtn_gradient_images_finite_differences_convolution(DD->ori_in1, grad0_x, grad1_y, 1, DD);

  /* Delete */
  mgtn_vector_delete(grad0_x);
  mgtn_vector_delete(grad0_y);
  mgtn_vector_delete(grad1_x);
  mgtn_vector_delete(grad1_y);

  /* Get filters for second mixed derivative */
  grad1_x = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_SECOND, DD->param.gradient_filter_size, 1, DD->param.interpolated_samples);
  grad1_y = mgtn_get_filter_gradient(DD->param.gradient_filter_type, D_SECOND, DD->param.gradient_filter_size, 1, DD->param.interpolated_samples);

  /* Compute gradients */
  DD->grad_t1_xy[0] = mgtn_gradient_images_finite_differences_convolution(DD->ori_in1, grad1_x, grad1_y, 1, DD);

  /* De-allocate memory */
  mgtn_vector_delete(grad1_x);
  mgtn_vector_delete(grad1_y);
}

/**
 *
 *  Compute the gradient images by restriction of gradient images computed
 *  at coarser levels. This function calls mgtn_gradient_images_restrict_gradient.
 *
 */

void mgtn_gradient_images_restriction(
    struct resources *DD)
{
  struct image *grad_x, *grad_y, *grad_x_prev, *grad_y_prev;
  int res, nrow, ncol, nrow2, ncol2;

  res = DD->curr_level;

  /* Gradient at time t0. Recall that gradients at t0 are computed with integer
   * pixel accuracy. 
   */

  nrow = GET_CURRENT_NROW(DD);
  ncol = GET_CURRENT_NCOL(DD);

  /* First order derivatives need to be restricted */

  grad_x_prev = DD->grad_t0_x[res-1];
  grad_y_prev = DD->grad_t0_y[res-1];

  grad_x = DD->grad_t0_x[res] = mgtn_image_alloc(NULL, nrow, ncol); 
  grad_y = DD->grad_t0_y[res] = mgtn_image_alloc(NULL, nrow, ncol); 

  mgtn_gradient_images_restrict_gradient(grad_x_prev, D_NOINTERP, grad_x, DD);
  mgtn_gradient_images_restrict_gradient(grad_y_prev, D_NOINTERP, grad_y, DD);

  /* Gradient at time t1. Recall that grdients at t1 are computed with subpixel
   * accuracy.
   */

  nrow2 = (nrow - 1) * DD->param.interpolated_samples + 1;
  ncol2 = (ncol - 1) * DD->param.interpolated_samples + 1;

  /* First order derivatives */

  grad_x_prev = DD->grad_t1_x[res-1];
  grad_y_prev = DD->grad_t1_y[res-1];

  grad_x = DD->grad_t1_x[res] = mgtn_image_alloc(NULL, nrow2, ncol2);
  grad_y = DD->grad_t1_y[res] = mgtn_image_alloc(NULL, nrow2, ncol2); 

  mgtn_gradient_images_restrict_gradient(grad_x_prev, D_INTERP, grad_x, DD);
  mgtn_gradient_images_restrict_gradient(grad_y_prev, D_INTERP, grad_y, DD);

  /* Second order derivative */

  if (DD->grad_t1_xx[0])
  {
    grad_x_prev = DD->grad_t1_xx[res-1];
    grad_y_prev = DD->grad_t1_yy[res-1];

    grad_x = DD->grad_t1_xx[res] = mgtn_image_alloc(NULL, nrow2, ncol2);
    grad_y = DD->grad_t1_yy[res] = mgtn_image_alloc(NULL, nrow2, ncol2);

    mgtn_gradient_images_restrict_gradient(grad_x_prev, D_INTERP, grad_x, DD);
    mgtn_gradient_images_restrict_gradient(grad_y_prev, D_INTERP, grad_y, DD);

    /* Mixed derivative */

    grad_x_prev = DD->grad_t1_xy[res-1];
    grad_x = DD->grad_t1_xy[res] = mgtn_image_alloc(NULL, nrow2, ncol2);

    mgtn_gradient_images_restrict_gradient(grad_x_prev, D_INTERP, grad_x, DD);
  }
}

/**
 *
 *  Compute the gradient images by restriction of gradient images computed at
 *  coarser levels. This function is called by
 *  mgtn_gradient_images_restriction. Note that there is an equivalent
 *  function called mgtn_image_restriction. The functions have been kept
 *  separately in order to be able to test what kind of restriction filter
 *  may be better in each case.
 *
 */

void mgtn_gradient_images_restrict_gradient(
    struct image *in, 
    int interpolated, 
    struct image *out, 
    struct resources *DD)
{
  int i, j, ncol, ncol2, nrow2;
  double *p_double_out, *p_double_tmp;

  struct vector *filter;
  struct image *tmp;

  /* Allocate temporal image */

  tmp = mgtn_image_alloc(NULL, in->nrow, in->ncol);
  mgtn_image_clear(tmp, 0.0);

  /* Filter image */

  filter = mgtn_get_filter_presmooth(D_CLASSICAL_RP, D_FILTER_RESTRICTION); 
  mgtn_image_separable_convolution(in, filter, filter, interpolated, D_FILTER_CLASSIC, tmp, DD);

  /* And now sample the image. Note that out has been previously
   * allocated. */

  ncol = tmp->ncol;

  ncol2 = out->ncol;
  nrow2 = out->nrow;

  /* Set samples in output image */

  for(i = 0; i < nrow2; i++)
  {
    p_double_out = out->gray + ncol2 * i;
    p_double_tmp = tmp->gray + ncol * (i << 1);

    for(j = 0; j < ncol2; j++, p_double_out++, p_double_tmp += 2)
      *p_double_out = *p_double_tmp;
  }

  /* Delete */

  mgtn_vector_delete(filter);
  mgtn_image_delete(tmp);
}

/**
 *
 *  Compute the gradient of an image via finite differences. The gradient is
 *  computed via a separable filter. The gradient is computed with subpixel
 *  accuracy. For this the input filters have to be interpolated with
 *  subpixel accuracy. See also function
 *  mgtn_gradient_images_finite_differences.
 *
 */

struct image *mgtn_gradient_images_finite_differences_convolution(
    struct image *in, 
    struct vector *filter_x, 
    struct vector *filter_y, 
    int flag,
    struct resources *DD)
{
  struct image *out, *tmp;

  double term_value, value;

  int interpolated_samples;
  int i, j, k_i, k_j, m_i, m_j, p_i, p_j, q_i, q_j, d_i, d_j;
  int mid_sample_i, mid_sample_j;
  int nrow, ncol, nrow_scaled, ncol_scaled;

  nrow = in->nrow;
  ncol = in->ncol;

  if (flag)
    interpolated_samples = DD->param.interpolated_samples;
  else
    interpolated_samples = 1;

  nrow_scaled = (nrow - 1) * interpolated_samples + 1;
  ncol_scaled = (ncol - 1) * interpolated_samples + 1;

  /* Allocate memory for output image */

  tmp = mgtn_image_alloc(NULL, nrow, ncol_scaled);
  out = mgtn_image_alloc(NULL, nrow_scaled, ncol_scaled);

  /* Apply filter */

  mid_sample_i = (filter_y->size / (2 * interpolated_samples));
  mid_sample_j = (filter_x->size / (2 * interpolated_samples));

  /* Filter in the x direction */

  for(i = 0; i < nrow ; i++)
    for(j = 0; j < ncol ; j++)
    {
      for(k_j = 0; k_j < interpolated_samples; k_j++)
      {
	/* Value will be stored in position (i, d_j) */

	d_j = interpolated_samples * j - k_j;

	/* Check if the value is going to be stored
	   in the image support */

	if ((d_j < 0) || (d_j >= ncol_scaled))
	  continue;

	/* Filter image in the x direction */

	value = 0.0;

	p_j = - mid_sample_j;

	for(m_j = k_j; m_j < filter_x->size; m_j += interpolated_samples, p_j++)
	{
	  q_j = j + p_j;

	  /* Border conditions */

	  if (q_j < 0)
	    q_j = 0;

	  if (q_j >= ncol)
	    q_j = ncol - 1;

	  /* Compute value and add to sum */

	  term_value = in->gray[i * ncol + q_j] * filter_x->values[m_j];

	  value += term_value;
	}

	tmp->gray[i * ncol_scaled + d_j] = value;

      }/*End for (k_i, k_j)*/
    }/*End for (i,j)*/

  /* Now filter in the y direction */

  for(i = 0; i < nrow ; i++)
    for(j = 0; j < ncol ; j++)
    {
      for(k_i = 0; k_i < interpolated_samples; k_i++)
	for(k_j = 0; k_j < interpolated_samples; k_j++)
	{
	  /* Value will be stored in position (d_i, d_j) */

	  d_i = interpolated_samples * i - k_i;
	  d_j = interpolated_samples * j - k_j;

	  /* Check if the value is going to be stored
	     in the image support */

	  if ((d_i < 0) || (d_i >= nrow_scaled))
	    continue;

	  if ((d_j < 0) || (d_j >= ncol_scaled))
	    continue;

	  /* Filter image in the y direction */

	  value = 0.0;

	  p_i = - mid_sample_i;

	  for(m_i = k_i; m_i < filter_y->size; m_i += interpolated_samples, p_i++)
	  {
	    q_i = i + p_i;

	    /* Border condition */

	    if (q_i < 0)
	      q_i = 0;

	    if (q_i >= nrow)
	      q_i = nrow - 1;

	    /* Compute value and add to sum */

	    term_value = tmp->gray[q_i * ncol_scaled + d_j] * filter_y->values[m_i];  

	    value += term_value;
	  }

	  out->gray[d_i * ncol_scaled + d_j] = (float) value;

	}/*End for (k_i, k_j)*/
    }/*End for (i,j)*/

  mgtn_image_delete(tmp);

  return out;
}

/**
 *
 *  Compute the gradient of an image via the method described by
 *  Horn-Schunk. This function is called if the linearized OFE is
 *  use. For convention I_x is stored in grad_t0_x, I_y is stored
 *  in grad_t0_y, and I_t is stored in grad_t1_x.
 *
 */

void mgtn_gradient_images_hornschunk(
    struct resources *DD)
{
  int i, nrow, ncol, npixels;
  double *p_target1, *p_target2, *p_target3, *p_target4, *p_in0, *p_in1;

  struct vector *filter_x, *filter_y;

  nrow = GET_CURRENT_NROW(DD);
  ncol = GET_CURRENT_NCOL(DD);

  npixels = nrow * ncol;

  /* Allocate memory */

  DD->grad_t0_x[0] = mgtn_image_alloc(NULL, nrow, ncol); 
  DD->grad_t0_y[0] = mgtn_image_alloc(NULL, nrow, ncol); 

  DD->grad_t1_x[0] = mgtn_image_alloc(NULL, nrow, ncol);
  DD->grad_t1_y[0] = mgtn_image_alloc(NULL, nrow, ncol);

  /* Compute in0 + in1 and in0 - in1 */

  p_target1 = DD->grad_t0_x[0]->gray;
  p_target2 = DD->grad_t0_y[0]->gray;
  p_target3 = DD->grad_t1_x[0]->gray;
  p_target4 = DD->grad_t1_y[0]->gray;

  p_in0  = DD->ori_in0->gray;
  p_in1  = DD->ori_in1->gray;

  for(i = 0; i < npixels; i++, p_target1++, p_target2++, p_target3++,
      p_target4++, p_in0++, p_in1++)
  {
    *p_target1 = *p_in0 + *p_in1;
    *p_target2 = *p_target1;
    *p_target3 = *p_in1 - *p_in0;
    *p_target4 = 0.0;
  }

  /* And now filter */

  filter_x = mgtn_vector_alloc(NULL, 2);
  filter_y = mgtn_vector_alloc(NULL, 2);

  filter_x->shift = -1;
  filter_y->shift = -1;

  /* Compute I_x, to be stored in grad_t0_x */

  filter_x->values[0] =  0.5;
  filter_x->values[1] = -0.5;

  filter_y->values[0] =  0.5;
  filter_y->values[1] =  0.5;

  mgtn_image_separable_convolution(DD->grad_t0_x[0], filter_x, filter_y, D_NOINTERP, D_FILTER_CLASSIC, DD->grad_t0_x[0], DD);

  /* Compute I_y, to be stored in grad_t0_y */

  filter_x->values[0] =  0.5;
  filter_x->values[1] =  0.5;

  filter_y->values[0] =  0.5;
  filter_y->values[1] = -0.5;

  mgtn_image_separable_convolution(DD->grad_t0_y[0], filter_x, filter_y, D_NOINTERP, D_FILTER_CLASSIC, DD->grad_t0_y[0], DD); 

  /* Compute I_t, to be stored in grad_t1_x */

  filter_x->values[0] =  0.5;
  filter_x->values[1] =  0.5;

  filter_y->values[0] =  0.5;
  filter_y->values[1] =  0.5;

  mgtn_image_separable_convolution(DD->grad_t1_x[0], filter_x, filter_y, D_NOINTERP, D_FILTER_CLASSIC, DD->grad_t1_x[0], DD); 

  /* Free memory */

  mgtn_vector_delete(filter_x);
  mgtn_vector_delete(filter_y);
}

/**
 *
 *  Interpolate the first order gradient for t0 or t1 at the given coordinates.
 *
 */

void mgtn_gradient_images_interpolate_first(
    int time,
    struct vector2 *coord,
    struct vector2 *grad,
    struct resources *DD)
{  
  int ncol, offset, x, y;

  struct vector2 coord_grad;
  struct image *image_x, *image_y;

  /*
   * For time t0 the gradient is computed with pixel precision
   * and evaluated only at integer pixel positions. For time
   * t1 the gradient is computed with subpixel precision and
   * can be evaluated at any non-integer pixel position. 
   */

  if (time == 0)
  {
    image_x = DD->grad_t0_x[DD->curr_level];
    image_y = DD->grad_t0_y[DD->curr_level];

    ncol   = image_x->ncol;

    x = coord->x;
    y = coord->y;
    offset = y * ncol + x; 

    grad->x = image_x->gray[offset];
    grad->y = image_y->gray[offset];
  }
  else
  {
    image_x = DD->grad_t1_x[DD->curr_level];
    image_y = DD->grad_t1_y[DD->curr_level];

    coord_grad.x = coord->x * DD->param.interpolated_samples;
    coord_grad.y = coord->y * DD->param.interpolated_samples;

    mgtn_image_interpolate_bilinear2(image_x, image_y, &coord_grad, &grad->x, &grad->y);
  }
}


/**
 *
 *  Interpolate the second order gradient for t1 at the given coordinates.
 *  This function is really not needed in the current implementation (i.e. the
 *  energy functionals), but has been included here to ease implementations of
 *  other types of energies.
 *
 */

void mgtn_gradient_images_interpolate_second(
    struct vector2 *coord,
    int type_grad,
    struct vector2 *grad,
    struct resources *DD)
{
  int res;

  struct vector2 coord_grad;
  struct image *image_x, *image_y;

  res = DD->curr_level;

  /* Get unormalized coordinates */

  coord_grad.x = coord->x * DD->param.interpolated_samples;
  coord_grad.y = coord->y * DD->param.interpolated_samples;

  /* Get image */

  switch (type_grad)
  {
    case D_SECOND:
      image_x = DD->grad_t1_xx[res];
      image_y = DD->grad_t1_yy[res];

      mgtn_image_interpolate_bilinear2(image_x, image_y, &coord_grad, &grad->x, &grad->y);
      break;

    case D_SECOND_MIXED:
      image_x = DD->grad_t1_xy[res];

      grad->x = mgtn_image_interpolate_bilinear(image_x, &coord_grad);
      grad->y = grad->x;

      break;
    default:
      error("type_grad in gradient interpolation not valid.\n");
  }
}

