//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 *
 * Perform bilinear interpolation of the image value
 *
 */

double mgtn_image_interpolate_bilinear(
  struct image *in, 
  struct vector2 *coord_in)
{
  int l, k, offset, ncol;
  double x, y, a, b, b_1, a_1, phi;
  double *x0, *x1, *x2, *x3;

  /* The 'floor' functions are not used here. It is 
     faster to asign directly than performing the 'floor' operator */

  x = coord_in->x;
  y = coord_in->y;

  l = x;
  k = y;

  a = x-l;
  b = y-k;

  ncol   = in->ncol;
  offset = k*ncol+l; 

  a_1 = 1.0 - a;
  b_1 = 1.0 - b;

  /* Bilinear interpolation */ 

  x0 = x1 = in->gray + offset;
  x1++;
  x2 = x3 = x0 + ncol;
  x3++;

  if ((!a) || (!b)) 
  {
    if ((!a) && (!b))
      phi = *x0;
    else
      if (!a)
	phi = b_1 * (*x0) + b * (*x2);
      else
	phi = a_1 * (*x0) + a * (*x1);
  } 
  else
    phi = 
      b_1 * (a_1 * (*x0) + a * (*x1)) + 
      b   * (a_1 * (*x2) + a * (*x3));

  return(phi);
}

/**
 *
 * Perform bilinear interpolation at two images. Rather than
 * calling twice the function mgtn_interpolate_bilinear, we
 * call this function to avoid repeating computations.
 *
 */

void mgtn_image_interpolate_bilinear2(
   struct image *in0, 
   struct image *in1, 
   struct vector2 *coord_in, 
   double *out0, 
   double *out1)
{
  int l, k, offset, ncol;
  double x, y, a, b, b_1, a_1;
  double x00, x10, x20, x30;
  double x01, x11, x21, x31;

  double *p;

  x = coord_in->x;
  y = coord_in->y;

  l = x;
  k = y;

  ncol   = in0->ncol;
  offset = k*ncol+l; 

  a = x-l;
  b = y-k;

  a_1 = 1.0 - a;
  b_1 = 1.0 - b;

  if ((!a) || (!b)) 
  {
    if ((!a) && (!b))
    {
      p   = in0->gray + offset;
      x00 = p[0];

      p = in1->gray + offset;
      x01 = p[0];

      *out0 = x00;
      *out1 = x01;
    }
    else
      if (!a)
      {
	p   = in0->gray + offset;

	x00 = p[0];
	x20 = p[ncol];

	p = in1->gray + offset;

	x01 = p[0];
	x21 = p[ncol];

	*out0 = b_1 * (x00) + b * (x20);
	*out1 = b_1 * (x01) + b * (x21);
      }
      else
      {
	p   = in0->gray + offset;

	x00 = p[0];
	x10 = p[1];

	p = in1->gray + offset;

	x01 = p[0];
	x11 = p[1];

	*out0 = a_1 * (x00) + a * (x10);
	*out1 = a_1 * (x01) + a * (x11);
      }
  } 
  else
  {
    p   = in0->gray + offset;

    x00 = p[0];
    x10 = p[1];
    x20 = p[ncol];
    x30 = p[ncol+1];

    p = in1->gray + offset;

    x01 = p[0];
    x11 = p[1];
    x21 = p[ncol];
    x31 = p[ncol+1];

    *out0 = 
      b_1 * (a_1 * (x00) + a * (x10)) + 
      b   * (a_1 * (x20) + a * (x30));

    *out1 = 
      b_1 * (a_1 * (x01) + a * (x11)) + 
      b   * (a_1 * (x21) + a * (x31));
  }
}


/**
 * 
 * Interpolate image t0 or t1 at given coordinates
 *
 */

double mgtn_image_interpolate(
  int time,
  struct vector2 *coord,
  struct resources *DD)
{
  struct image *image;

  int x, y, ncol, offset;
  double value;

  if (time == 0)
  {
    image = DD->in0[DD->curr_level];

    ncol   = image->ncol;

    x = coord->x;
    y = coord->y;
    offset = y * ncol + x; 

    value = image->gray[offset];
  }
  else
  {
    image = DD->in1[DD->curr_level];

    value = mgtn_image_interpolate_bilinear(image, coord);
  }

  return value;
}

/**
 * 
 *  Perform image restriction. The image is assumed not be be computed with
 *  subpixel accuracy (option D_NOINTERP within the function). Note that
 *  there is an equivalent function for the gradients,
 *  mgtn_gradient_images_restrict_gradient.  The functions have been kept
 *  separated in order to be able to test what kind of restriction filter may
 *  be better in each case.
 *
 */

void mgtn_image_restriction(
  struct image *in, 
  struct image *out, 
  struct resources *DD)
{
  int i, j, ncol, ncol2, nrow2;
  double *p_double_out, *p_double_tmp;

  struct vector *filter;
  struct image *tmp;

  /* Allocate temporal image */

  tmp = mgtn_image_alloc(NULL, in->nrow, in->ncol);
  mgtn_image_clear(tmp, 0.0);

  /* Filter image */

  filter = mgtn_get_filter_presmooth(D_CLASSICAL_RP, D_FILTER_RESTRICTION); 
  mgtn_image_separable_convolution(in, filter, filter, D_NOINTERP, D_FILTER_CLASSIC, tmp, DD);

  /* And now sample the image. Note that out has been previously
   * allocated. */

  ncol = tmp->ncol;

  ncol2 = out->ncol;
  nrow2 = out->nrow;

  /* Set samples in output image */

  for(i = 0; i < nrow2; i++)
  {
    p_double_out = out->gray + ncol2 * i;
    p_double_tmp = tmp->gray + ncol * (i << 1);

    for(j = 0; j < ncol2; j++, p_double_out++, p_double_tmp += 2)
      *p_double_out = *p_double_tmp;
  }

  /* Delete */

  mgtn_vector_delete(filter);
  mgtn_image_delete(tmp);
}

/**
 * 
 *  Perform image prolongation. 
 *
 */

void mgtn_image_prolongation(
  struct image *in, 
  struct image *out,
  struct resources *DD)
{
  int i, j, ncol, nrow;
  double *p_double_out, *p_double_in;

  struct vector *filter;

  ncol = out->ncol;
  nrow = out->nrow;

  /* Initialize out to zero */

  mgtn_image_clear(out, 0.0);

  /* Set samples in output image */

  p_double_out = out->gray;
  p_double_in  = in->gray;

  for(i = 0; i < nrow; i += 2, p_double_out += ncol)
  {
    for(j = 0; j < ncol; j += 2, p_double_in++, p_double_out += 2)
      *p_double_out = *p_double_in;

    if (ODD(ncol))
      p_double_out--;
  }

  filter = mgtn_get_filter_presmooth(D_PRESMOOTH_11, D_FILTER_PROLONGATION);
  mgtn_image_separable_convolution(out, filter, filter, D_NOINTERP, D_FILTER_PROLONGATION, out, DD);

  mgtn_vector_delete(filter);
}

/**
 * 
 *  Perform restriction for a pair of (x,y) images. 
 *
 */

void mgtn_image_restriction2(
  struct image *image_x,
  struct image *image_y,
  struct image *image_x_new,
  struct image *image_y_new,
  struct resources *DD)
{
  /* Scale images */

  mgtn_image_restriction(image_x, image_x_new, DD);
  mgtn_image_restriction(image_y, image_y_new, DD);
}

/**
 * 
 *  Perform prolongation for a pair of (x,y) images. 
 *
 */

void mgtn_image_prolongation2(
  struct image *image_x,
  struct image *image_y,
  struct image *image_x_new,
  struct image *image_y_new,
  struct resources *DD)
{
  /* Scale images */

  mgtn_image_prolongation(image_x, image_x_new, DD);
  mgtn_image_prolongation(image_y, image_y_new, DD);
}


/**
 * 
 *  Perform convolution of an image with a separable filter. The input
 *  image can be the output image. Output image has to be allocated.
 *  The parameter flag is used to indicate what type of filtering is
 *  applied: it may be D_FILTER_CLASSIC, in which a classical filtering
 *  over all pixels is applied, of D_FILTER_PROLONGATION, in which we
 *  assume that the image is an already zoomed image in which we have
 *  put the image values with neighbouring zero pixels (the prolongation
 *  is performed as is done in signal processing).
 *
 */

void mgtn_image_separable_convolution(
  struct image *in, 
  struct vector *xker, 
  struct vector *yker,
  int interpolated,
  int flag,
  struct image *out,
  struct resources *DD)
{
  int interpolated_samples;
  int nx, nx1, ny, ny1, x, y, org, i, s;
  int min_x, min_y, max_x, max_y;

  double sum;
  struct image *tmp;

  interpolated_samples = 1;

  if (interpolated == D_INTERP)
    interpolated_samples = DD->param.interpolated_samples;
  else if (interpolated == D_NOINTERP)
    interpolated_samples = 1;
  else 
    error("Interpolated value not valid.\n");

  /* Initialize temporal image. Output image shoud be allocated. */

  nx = in->ncol; nx1 = nx - 1;
  ny = in->nrow; ny1 = ny - 1;

  min_x = 0;
  min_y = 0;

  max_x = nx1;
  max_y = ny1;

  /* Initialize */

  tmp = mgtn_image_alloc(NULL,ny,nx);
  mgtn_image_clear(tmp, 0.0);

  /* convolution along x axis */

  org = - (int) xker->shift;

  for (y = min_y; y <= max_y; y++)
  {
    if ((flag == D_FILTER_PROLONGATION) && (ODD(y)))
      continue;

    for (x = min_x;x <= max_x; x++) 
    {
      sum = 0.0;

      for (i=xker->size;i--;) 
      {
	s = x-(i-org)*interpolated_samples;

	/* Border condition */

	if (s < min_x) s = min_x;
	if (s > max_x)
	{
	  if (flag == D_FILTER_PROLONGATION)
	  {
	    {
	      if (ODD(max_x))
		s = max_x - 1;
	      else
		s = max_x;
	    }
	  }
	  else
	    s = max_x;
	}

	/* Update */

	sum += xker->values[i]*in->gray[y*nx+s];
      }

      tmp->gray[y*nx+x] = sum;
    }
  }

  /* convolution along y axis */

  org = -(int)yker->shift;

  for (y = min_y;y <= max_y; y++)
  {
    for (x = min_x; x <= max_x; x++) 
    {
      sum = 0.0;

      for (i=yker->size;i--;) 
      {
	s = y-(i-org)*interpolated_samples;

	/* Border condition */

	if (s < min_y) s = min_y;
	if (s > max_y) 
	{
	  if (flag == D_FILTER_PROLONGATION)
	  {
	    {
	      if (ODD(max_y))
		s = max_y - 1;
	      else
		s = max_y;
	    }
	  }
	  else
	    s = max_y;
	}

	/* Update */

	sum += yker->values[i]*tmp->gray[s*nx+x];
      }

      out->gray[y*nx+x] = sum;
    }
  }

  /* Free memory */

  mgtn_image_delete(tmp);
}

/**
 *
 *  Prefilters original input images
 *
 */

void mgtn_image_prefilter_images(
    struct image *in0,
    struct image *in1,
    struct resources *DD)
{
  struct vector *prefilter;

  prefilter = mgtn_get_filter_prefilter(DD->param.prefilter_size, DD->param.prefilter_sigma);

  mgtn_image_separable_convolution(in0, prefilter, prefilter, D_NOINTERP, D_FILTER_CLASSIC, in0, DD);
  mgtn_image_separable_convolution(in1, prefilter, prefilter, D_NOINTERP, D_FILTER_CLASSIC, in1, DD);

  mgtn_vector_delete(prefilter);
}

/**
 *
 *  Computes the pyramid for an image using restriction. The
 *  vector storing the output images should be previously
 *  allocated (but not the images itself).
 *
 */

void mgtn_image_compute_pyramid_image(
    struct image *in_image, 
    struct image **out_images, 
    struct resources *DD)
{
  int i;

  struct image *image, *image_restricted;

  /* For the finest level image just copy it */

  image = mgtn_image_alloc(NULL, in_image->nrow, in_image->ncol);
  mgtn_image_copy(in_image, image);

  /* Assign to finest level */

  out_images[0] = image; 

  /* Go through all the levels */ 

  for(i = 1; i <= DD->param.max_level; i++)
  {
    /* Allocate memory for the new image */

    image_restricted = mgtn_image_alloc(NULL, DD->nrow[i], DD->ncol[i]);

    /* Restrict image from the previous level */

    mgtn_image_restriction(image, image_restricted, DD);

    /* Assign to pyramid */

    out_images[i] = image_restricted;

    /* Prepare data for next iteration */

    image = image_restricted;
    image_restricted = NULL;
  }
}

/**
 *
 *  Computes the pyramid for the two input images at t0 and
 *  t1. This function calls mgtn_image_compute_pyramid_image.
 *
 */

void mgtn_image_compute_pyramid(
    struct resources *DD)
{
  int i;

  mgtn_image_compute_pyramid_image(DD->ori_in0, DD->in0, DD);
  mgtn_image_compute_pyramid_image(DD->ori_in1, DD->in1, DD);

  printMsg("Resolution of pyramid images: ");

  for(i = 0; i <= DD->param.max_level; i++)
    printMsg("%d x %d, ", DD->in0[i]->ncol, DD->in0[i]->nrow);

  printMsg("\n");
}

