//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include <stdio.h>
#include <stdlib.h>

#include "mgtn.h"
#include "cmdline.h"

/*
 * If variable quiet is set to true, application does not output
 * messages to screen. By default it is set to false.
 */

int quiet = 0;

/**
 *
 *  Main function 
 *
 */

int main(int argc, char **argv)
{
  FILE *fp;
  char *fname;
  struct arguments *args = process_arguments(USE, argc, argv);

  /* The structure resources holds all the data that is needed 
     for computing the optical flow. This structure is passed
     to the functions to as to avoid passing a large number of
     parameters */

  struct resources DD;

  /* The two images of which the optical flow has to be 
     estimated */

  struct image *in0, *in1;

  /* Read input images */

  in0 = mgtn_pgm_read_image(get_str(args, "inT0"));
  in1 = mgtn_pgm_read_image(get_str(args, "inT1"));

  /* Check if images have the same size */

  if ((in0->ncol != in1->ncol) || (in0->nrow != in1->nrow))
    error("ERROR: input image sizes are not the same.\n");

  /* Set input images */

  mgtn_set_input_images(in0, in1, &DD);

  /* Has the application be quiet ? */

  quiet = get_int(args, "quiet");

  /* Read configuration file */

  fname = get_str(args, "conffile");

  fp = mgtn_open_configuration_file(fname);

  if (!fname)
    mgtn_generate_configuration_file(fp, get_int(args, "model"), get_int(args, "scheme"), 
      get_double(args, "alpha"), get_int(args, "vcycles"));
  else
    printMsg("Configuration file specified. Remaining configuration command-line arguments are ignored.\n");

  mgtn_read_configuration_file(fp, &DD);

  mgtn_close_configuration_file(fp);

  /* Initialize remaining resources */ 

  mgtn_resources_initialize(&DD);

  /* Read ground truth if available */

  fname = get_str(args, "gtruth");
  if ((fname) && (quiet))
     fprintf(stderr, "Warning: reading ground truth but application is quiet.\n");
  
  if (fname)
    mgtn_groundtruth_read(fname, &DD);

  /* Execute dense motion field */

  dense_motion_field(&DD);

  /* Done. Print out some output */

  if  (!quiet)
  {
    printMsg("\nDisplaced frame difference error: %f\n", mgtn_dfd(5, &DD));

    if (DD.have_groundtruth_flow)
    {
      double val1, val2;

      val1 = mgtn_end_point_error(DD.u[0], DD.v[0], DD.groundtruth_flow_u, DD.groundtruth_flow_v);
      val2 = mgtn_average_angular_error(DD.u[0], DD.v[0], DD.groundtruth_flow_u, DD.groundtruth_flow_v);

      printMsg("End point error (EPE): %7.5f\n", val1);
      printMsg("Average angular error (AAE): %7.5f degrees\n", val2);
    }
  }

  /* Dump flow  in a file */
  mgtn_flow_write(get_str(args, "outF"), &DD);

  /* Dump backwards compensated image */
  fname = get_str(args, "backcomp");
  if (fname)
    mgtn_backcomp_write(fname, &DD);

  /* Delete the structure resources */
  mgtn_resources_delete(&DD);

  /* Delete input images */
  mgtn_image_delete(in0);
  mgtn_image_delete(in1);

  /* Free arguments */
  free(args);

  /* Done */
  printMsg("Done.\n");

  return 0;
}



