//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#ifndef MGTN_H
#define MGTN_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <stdarg.h>
#include <float.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

/*
 *  Threshold for the input parameters. These values seem to work pretty
 *  well.
 *
 */

#define MIN_RESOLUTION_LEVELS      1
#define MAX_RESOLUTION_LEVELS      8

#define MIN_ALPHA                  10.0
#define MAX_ALPHA                  10000.0

#define MIN_UNILEVEL_ITERATIONS    25
#define MAX_UNILEVEL_ITERATIONS    1000

#define MIN_MR_ITERATIONS          5
#define MAX_MR_ITERATIONS          100

#define MIN_V_CYCLES               1
#define MAX_V_CYCLES               25

/* 
 * Uncomment if your computer has not much memory (< 2 GBytes) or you use big
 * images. A lot of memory may be used due to the interpolation factor value.
 * Indeed, if the image has a size W x H (width x height), the gradient will
 * have a size of (F * W) x (F * H), where F is the interpolation factor. 
 *
 * If commented all the gradients will be stored in memory. Otherwise gradient
 * will be temporally stored on disc. The the temporal files will be
 * automatically removed when the application ends.
 *
 */

//#define LOW_MEMORY_COMPUTER

/*
 * Basic structures needed by the algorithm 
 */

/* The image structure */

struct image {
  int nrow;         /* Number of rows (dy) */
  int ncol;         /* Number of columns (dx) */
  double *gray;     /* The Gray level plane (may be NULL) */
};

/* A vector made up of 2 components */

struct vector2 {
  double x;
  double y;
};

/* A vector with a shift (used for filtering) */

struct vector {
  int size;
  int shift;
  double *values;
};

/*
 * Parameters to be read from configuration file parameters.cfg. The order
 * in which they are written in this structure is the order in which
 * they are expected to be in the configuration file. See function
 * mgtn_read_configuration_file.
 */

struct parameters {

  int scheme;                   /* Scheme to be used: multiresoltion or multigrid */
  int max_level;                /* Maximum level. The 0 is the finest level */

  int functional_data_type;     /* Type of functional to apply */
  int functional_regul_type;    /* Type of regularizer */

  double alpha;                 /* Weight data term-regularizer */
  double mu;                    /* Mu for motion regularization */
  double outlier;               /* Value for outlier used for thredholding energy */

  int prefilter_size;           /* Size for the length of the prefilter */
  double prefilter_sigma;       /* Sigma, a gaussian filter is used */

  int gradient_filter_type;     /* Filter type of image gradient computation */
  int gradient_filter_size;     /* Stencil size */
  int interpolated_samples;     /* Number of interpolated samples */

  double ftol, xtol, pgtol;     /* ftol, xtol and pgtol */
  int maxit_tn_internal;        /* Maximum number of internal iterations */

  int maxit_multiresolution;    /* Maximum number of iterations for each level in multiresolution */
  int maxit_coarse_multigrid;   /* Maximum number of iterations in multigrid's coarsest level */

  int type_multigrid;           /* Type of multigrid: V or W cycle */
  int check_descent;            /* For postsmooth: check if correction direction already provides descent */
  double beta_constraint;       /* Used to constrain search */ 
  double a_presmooth_cond;      /* Used to check presmoothing condition */
  double e_presmooth_cond;      /* Used to check presmoothing condition */

  int *maxit_v_cycle;       /* Number of v-cycles to perform at each level */
  int *maxit_presmoothing;  /* Maximum number of iterations for the presmoothing at each level */
  int *maxit_postsmoothing; /* Maximum number of iterations for the postsmoothing at each level */
}; 

/*
 * The resources of the application. This structure contains all the data
 * the application needs to run properly. With this struture we avoid 
 * passing a long number of arguments to the function. Rather than passing
 * each variable as an argument, we pass only this structure which contains
 * all the needed information.
 */

struct resources {

  struct parameters param;

  /* Used to compute elapsed time. The elapsed time does not
   * take into account loading of data from disc */

  int t_active;
  clock_t t_elapsed, t_last;

  int *nfeval, *ngeval;

  /* Some needed variables */

  int curr_level;

  /* Value related to h */

  double inverse_h;

  /* The nrow and col for the input images along the pyramid */

  int *nrow;
  int *ncol;

  /* The original images */

  struct image *ori_in0;
  struct image *ori_in1;

  /* The pyramid associated to the input images. This structure
   * should be interpreted as a vector of pointers to images. */

  struct image **in0;
  struct image **in1;

  /* The gradient images. It should be interpreted as a vector of pointers to
   * images. */
  
  struct image **grad_t0_x;
  struct image **grad_t0_y;

  struct image **grad_t1_x;
  struct image **grad_t1_y;

  struct image **grad_t1_xx;
  struct image **grad_t1_yy;
  struct image **grad_t1_xy;

  /* Flow images. It should be interpreted as a vector of pointers to images */

  struct image **u;
  struct image **v;

  int have_groundtruth_flow;
  struct image *groundtruth_flow_u;
  struct image *groundtruth_flow_v;

  /* Gradient of the energy. It should be interpreted as a vector of pointers
   * to images */

  struct image **grad_E_u; 
  struct image **grad_E_v;
 
  /* 
   * The members below are related to the multigrid method 
   */

  /* Flags to manage the multigrid technique */

  int in_multigrid;
  int in_presmoothing;

  /* Variables to manage the resolution levels to analyze */

  int curr_level_cycle;
  int *cycle;

  /* Residue. It should be interpreted as a vector of pointers to images. */

  struct image **residue_u;
  struct image **residue_v;

  /* Restricted flow images. It should be interpreted as a vector of pointers
   * to images. */

  struct image **restricted_u;
  struct image **restricted_v;

  /* Temporal images for the multigrid process.  */

  struct image **tmp_u;
  struct image **tmp_v;
};

/* Please DO NOT change the values defined below unless you know 
 * what you are doing */

/* Schemes */

#define SCHEME_MULTIRESOLUTION      1
#define SCHEME_FMG                  2 // FMG = Full MultiGrid

/* Model 1, 2, 3 and 4 according to the paper */

#define MODEL_LINEAR_QUADRATIC      1
#define MODEL_NON_LINEAR_QUADRATIC  2
#define MODEL_LINEAR_TV             3
#define MODEL_NON_LINEAR_TV         4 

/* Linear and non-linear data term */

#define FUNCTIONAL_DATA_NON_LINEAR  1   
#define FUNCTIONAL_DATA_LINEAR      2

/* Quadratic and non-quadratic regularization term */

#define FUNCTIONAL_REGUL_QUADRATIC  1
#define FUNCTIONAL_REGUL_TV         2

/* Image gradient computation */

#define FILTER_SIMONCELLI    1
#define FILTER_GAUSS         2
#define FILTER_OFE_HS        3 // Optical Flow Equation, Horn & Schunk

/* Multigrid: V or W cycle */

#define V_CYCLE              1
#define W_CYCLE              2

/* Multigrid: should multigrid try to check if correction step already
 * reduces energy function ? If yes, update is accepted. If not, a
 * line search is performed even if correction step already reduces
 * energy function.
 */

#define MG_CHECK_DESCENT_NO   0  
#define MG_CHECK_DESCENT_YES  1 

/* Motion derivatives (forward/backward differences) */

#define VXf              0
#define VYf              1
#define UXf              2
#define UYf              3
#define VXb              4
#define VYb              5
#define UXb              6
#define UYb              7

/* Initial guess for bracketing in line search */

#define STEP_LINSRCH              1.0

/* Macros related to descent */

#define MGTN_TNC_MAXITER        0
#define MGTN_TNC_CONVERGED      1
#define MGTN_TNC_USERABORT      2
#define MGTN_TNC_PRECOND        3
#define MGTN_TNC_CORRNOTDESC    4

/* General macros */

#define MIN(x,y)          	 ((x)<(y)?(x):(y))
#define MAX(x,y)          	 ((x)<(y)?(y):(x))

#define ODD(x)			 ((x) & 0x01)
#define EVEN(x)                  (!((x) & 0x01))

#define SQR(x)            	 ((x)*(x))

#define GET_CURRENT_IMAGE(DD,VAR)      (DD)->VAR[(DD)->curr_level]
#define GET_COARSER_IMAGE(DD,VAR)      (DD)->VAR[(DD)->curr_level+1]
#define GET_FINER_IMAGE(DD,VAR)        (DD)->VAR[(DD)->curr_level-1]

#define GET_CURRENT_NROW(DD)           (DD)->nrow[(DD)->curr_level]
#define GET_CURRENT_NCOL(DD)           (DD)->ncol[(DD)->curr_level]

/* Maximum number of function evaluations. This value should be high
 * enough so that is does not affect the program.*/

#define MAXNFEVAL                   200000

/* A magick key used for the gradient files used in this algorithm */

#define MAGIC_KEY 0xAABBCCDD

/* For gradient computation */

#define D_NONE                     0
#define D_FIRST                    1
#define D_SECOND                   2
#define D_SECOND_MIXED             3

/* For image and gradient restriction */

#define D_NOINTERP                 0
#define D_INTERP                   1

/* For filtering using separable convolution */

#define D_FILTER_CLASSIC            0
#define D_FILTER_RESTRICTION        1
#define D_FILTER_PROLONGATION       2

/* For mgtn_get_filter_presmooth  */

#define D_CLASSICAL_RP             0
#define D_PRESMOOTH_5              1
#define D_PRESMOOTH_7              2
#define D_PRESMOOTH_9              3
#define D_PRESMOOTH_11             4
#define D_INTERPOLATE_7            5

/* Prototypes */

/* dmf.c */
void dense_motion_field(struct resources *DD);
/* eval_functional.c */
double mgtn_functional_evaluation_multigrid(struct image *u, struct image *v, struct image *residue_u, struct image *residue_v, struct resources *DD);
double mgtn_functional_evaluation_residue(struct image *u, struct image *v, struct image *residue_u, struct image *residue_v);
double mgtn_functional_evaluation(struct image *u, struct image *v, struct resources *DD);
/* eval_gradient.c */
void mgtn_gradient_evaluation_multigrid(struct image *u, struct image *v, struct image *residue_u, struct image *residue_v, struct image *grad_E_u, struct image *grad_E_v, struct resources *DD);
void mgtn_gradient_evaluation_residue(struct image *grad_E_u, struct image *grad_E_v, struct image *residue_u, struct image *residue_v);
void mgtn_gradient_evaluation(struct image *u, struct image *v, struct image *grad_E_u, struct image *grad_E_v, struct resources *DD);
void mgtn_gradient_evaluation_data(struct vector2 *coord_pix, struct image *u, struct image *v, double *grad_data_u, double *grad_data_v, struct resources *DD);
void mgtn_gradient_evaluation_reg(struct vector2 *coord_pix, struct image *u, struct image *v, double *grad_reg_u, double *grad_reg_v, struct resources *DD);
/* eval_utils.c */
int mgtn_get_motion_derivatives(double *p_u, double *p_v, double motder[4], struct resources *DD);
void mgtn_set_border_conditions_uv(struct image *u, struct image *v);
void mgtn_set_border_conditions_gradient(struct image *grad_u, struct image *grad_v);
/* image.c */
struct image *mgtn_image_new(void);
struct image *mgtn_image_alloc(struct image *image, int nrow, int ncol);
void mgtn_image_delete(struct image *image);
void mgtn_image_copy(struct image *in, struct image *out);
void mgtn_image_clear(struct image *im, double value);
/* image_gradient.c */
void mgtn_gradient_images_initialize(struct resources *DD);
void mgtn_gradient_images_delete(struct resources *DD);
void mgtn_gradient_images_finite_differences(struct resources *DD);
void mgtn_gradient_images_restriction(struct resources *DD);
void mgtn_gradient_images_restrict_gradient(struct image *in, int interpolated, struct image *out, struct resources *DD);
struct image *mgtn_gradient_images_finite_differences_convolution(struct image *in, struct vector *filter_x, struct vector *filter_y, int flag, struct resources *DD);
void mgtn_gradient_images_hornschunk(struct resources *DD);
void mgtn_gradient_images_interpolate_first(int time, struct vector2 *coord, struct vector2 *grad, struct resources *DD);
void mgtn_gradient_images_interpolate_second(struct vector2 *coord, int type_grad, struct vector2 *grad, struct resources *DD);
/* image_io.c */
struct image *mgtn_pgm_read_image(char * name);
void mgtn_pgm_write_image(struct image *image, char * name);
/* image_utils.c */
double mgtn_image_interpolate_bilinear(struct image *in, struct vector2 *coord_in);
void mgtn_image_interpolate_bilinear2(struct image *in0, struct image *in1, struct vector2 *coord_in, double *out0, double *out1);
double mgtn_image_interpolate(int time, struct vector2 *coord, struct resources *DD);
void mgtn_image_restriction(struct image *in, struct image *out, struct resources *DD);
void mgtn_image_prolongation(struct image *in, struct image *out, struct resources *DD);
void mgtn_image_restriction2(struct image *image_x, struct image *image_y, struct image *image_x_new, struct image *image_y_new, struct resources *DD);
void mgtn_image_prolongation2(struct image *image_x, struct image *image_y, struct image *image_x_new, struct image *image_y_new, struct resources *DD);
void mgtn_image_separable_convolution(struct image *in, struct vector *xker, struct vector *yker, int interpolated, int flag, struct image *out, struct resources *DD);
void mgtn_image_prefilter_images(struct image *in0, struct image *in1, struct resources *DD);
void mgtn_image_compute_pyramid_image(struct image *in_image, struct image **out_images, struct resources *DD);
void mgtn_image_compute_pyramid(struct resources *DD);
/* multigrid.c */
void mgtn_initialize_w_cycle(struct resources *DD);
void mgtn_generate_w_cycle(struct resources *DD);
void mgtn_initialize_v_cycle(struct resources *DD);
void mgtn_generate_v_cycle(struct resources *DD);
int mgtn_multigrid(struct resources *DD);
/* resources.c */
void mgtn_set_input_images(struct image *in0, struct image *in1, struct resources *DD);
void mgtn_resources_initialize(struct resources *DD);
void mgtn_resources_init_presmooth(struct resources *DD);
void mgtn_resources_init_postsmooth(struct resources *DD);
void mgtn_resources_delete_presmooth(struct resources *DD);
void mgtn_resources_delete_postsmooth(struct resources *DD);
void mgtn_resources_delete(struct resources *DD);
/* tnc_utils.c */
int mgtn_tnc(double *f, struct image *u, struct image *v, struct image *g_u, struct image *g_v, int maxiter, int relative_res, struct resources *DD);
int mgtn_tnc_function(double *x, double *f, double *g, void *state);
/* utils.c */
void mgtn_valgrind_dummy_function(struct image *in);
void *xmalloc(size_t size);
void error(char *msg);
void printMsg( const char* pFormat, ... );
void mgtn_initialize_nrow_ncol(int nrow, int ncol, struct resources *DD);
void mgtn_split_vector_uv(double *x, struct image *u, struct image *v);
void mgtn_merge_vector_uv(struct image *u, struct image *v, double *x);
void mgtn_flow_initialize(struct resources *DD);
int mgtn_fourier_power_of_two(int n);
void mgtn_fourier_fft1d(struct vector *Xr, struct vector *Xi, struct vector *Yr, struct vector *Yi, int inverse);
void mgtn_fourier_nonintegershift(struct vector *filter, double shift, struct vector *filter_shift);
struct vector *mgtn_get_filter_prefilter(int length, double sigma);
struct vector *mgtn_get_filter_gradient(int filter_type, int derivative_order, int length, int index, int interpolated_samples);
struct vector *mgtn_get_filter_presmooth(int filter_id, int type_filter);
double mgtn_dfd(int border, struct resources *DD);
double mgtn_gradient_norm(struct image *u, struct image *v);
double mgtn_end_point_error(struct image *u1, struct image *v1, struct image *u2, struct image *v2);
double mgtn_average_angular_error(struct image *u1, struct image *v1, struct image *u2, struct image *v2);
/* utils_io.c */
struct image *mgtn_read_image(const char *filename);
#ifdef LOW_MEMORY_COMPUTER
int mgtn_read_gradient_images(struct resources *DD);
void mgtn_write_gradient_images(struct resources *DD);
#endif
FILE *mgtn_open_configuration_file(char *fname);
void mgtn_close_configuration_file(FILE *fp);
void mgtn_generate_configuration_file(FILE *fp, int model, int scheme, double alpha, int vcycles);
void mgtn_read_configuration_file(FILE *fp, struct resources *DD);
void mgtn_groundtruth_read(char *fname, struct resources *DD);
void mgtn_flow_write(char *fname, struct resources *DD);
void mgtn_backcomp_write(char *fname, struct resources *DD);
/* vector.c */
struct vector *mgtn_vector_new(void);
struct vector *mgtn_vector_alloc(struct vector *vector, int N);
void mgtn_vector_delete(struct vector *vector);
void mgtn_vector_clear(struct vector *vector, double value);

#endif


