//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (C) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>
//               Mostafa Kalmoun <kalmoun@ensa.ac.ma>

#include <stdio.h>
#include "mgtn.h"
#include "tnc.h"

/*
 * Output strings. Should coincide with the codes used
 * in mgtn.h
 */

static char *mgtn_rc_string[5] =
{
  "maxiter reached",
  "converged",
  "user aborted",
  "presmoothing condition satisfied",
};

/**
 *
 *  This funcion (together with mgtn_generate_w_cycle)
 *  to set up the order in which the resolution levels
 *  have to be used in a W-cycle.
 *
 */

void mgtn_initialize_w_cycle(
  struct resources *DD)
{
  int i, k, length;

  length = 1;

  k = DD->param.max_level;

  for(i = k; i > 0; i--)
    length = length * 2 + 1;

  DD->cycle = (int *) malloc(sizeof(int) * (length + 1));

  for(i = 0; i <= length; i++)
    DD->cycle[i] = -1;
}

/**
 *
 *  This funcion (together with mgtn_initialize_w_cycle)
 *  to set up the order in which the resolution levels
 *  have to be used in a W-cycle.
 *
 */

void mgtn_generate_w_cycle(
  struct resources *DD)
{
  int i, j, k, curr, next;

  curr = DD->curr_level;
  k = DD->param.max_level;

  if (curr == k)
  {
    DD->cycle[0] = k;
    return;
  }

  i = 0;

  do {
    next = DD->cycle[i];
    DD->cycle[i] = curr;
    curr = next;
    i++;
  } while (next != -1);

  /* And repeat in reverse order */

  i--;
  for(j = 0; j < i; j++)
    DD->cycle[i + j + 1] = DD->cycle[i - j - 1];
}

/**
 *
 *  This funcion (together with mgtn_generate_v_cycle)
 *  to set up the order in which the resolution levels
 *  have to be used in a W-cycle.
 *
 */

void mgtn_initialize_v_cycle(
  struct resources *DD)
{
  int i, k, length;

  length = 1;

  k = DD->param.max_level;

  for(i = k; i > 0; i--)
    length = length + 2;

  DD->cycle = (int *) malloc(sizeof(int) * (length + 1));

  for(i = 0; i <= length; i++)
    DD->cycle[i] = -1;
}

/**
 *
 *  This funcion (together with mgtn_initialize_v_cycle)
 *  to set up the order in which the resolution levels
 *  have to be used in a W-cycle.
 *
 */

void mgtn_generate_v_cycle(
  struct resources *DD)
{
  int i, k, curr, next;

  curr = DD->curr_level;
  k = DD->param.max_level;

  if (curr == k)
  {
    DD->cycle[0] = k;
    return;
  }

  i = 0;

  do {
    next = DD->cycle[i];
    DD->cycle[i] = curr;
    curr = next;
    i++;
  } while (next != -1);

  DD->cycle[i] = DD->curr_level;
}

/**
 *
 * This function is the heart of th multigrid. Here the multigrid
 * is executed. This function only performs a V or W cycle for a
 * given level. The control of this function is performed from the
 * main function.
 *
 */

int mgtn_multigrid(
  struct resources *DD)
{
  struct image *u, *v, *grad_E_u, *grad_E_v;

  int rc, i, res, res0, relative_res, nb_pixels, index;

  double *lastf, *p_tmp, *p_d1, *p_d2, *p_d3;

  rc    = 0;
  index = 0;

  /* 
   * The vector lastf is used to store, for a given resolution level,
   * the last value of the energy evaluation. Thus, when going up and down
   * in the hierarchy, there is no need to re-evaluate the energy when 
   * comming back from another level 
   */

  lastf = (double *) xmalloc(sizeof(double) * (DD->param.max_level + 1));

  for(i = 0; i <= DD->param.max_level; i++)
    lastf[i] = 0.0;

  /*
   * The vector w-cycle stores the order in which the algorithm has to
   * go through the levels.
   */

  res = res0 = DD->cycle[index];

  /*
   * Check that everything is ok
   */

  if (DD->curr_level != res) 
    error("multigrid: DD->curr_level != res, this should not happen.\n");

 /*
  * We evaluate first time the function outside the TNC library.
  * Thus, we try to minimize the number of times the function is
  * evaluated for the multigrid. 
  */

  u = GET_CURRENT_IMAGE(DD, u);
  v = GET_CURRENT_IMAGE(DD, v);

  grad_E_u = GET_CURRENT_IMAGE(DD, grad_E_u);
  grad_E_v = GET_CURRENT_IMAGE(DD, grad_E_v);

  lastf[res] = mgtn_functional_evaluation(u, v, DD);
  mgtn_gradient_evaluation(u, v, grad_E_u, grad_E_v, DD);

  DD->nfeval[res]++;
  DD->ngeval[res]++;

  /* 
   *  Here does the multigrid code begin. The DD->cycle indicates
   *  the order in which the levels have to be analysed. Eg, cycle
   *  may contain something like
   *    0 1 2 3 4 5 4 3 2 1 0 -1
   *  which means that the cycle begins at level 0, then goes to
   *  level 1, then to level 2, until level 5 is reached. Afterwards
   *  we go back until level 0. The -1 indicates the end of the level.
   *  The generation of the cycle vectors is performed by the 
   *  generate_v_cycle and generate_w_cycle functions.
   */

  while (DD->cycle[index] != -1) 
  {
    res          = DD->cycle[index];
    relative_res = DD->curr_level_cycle = res - res0; 

    u = GET_CURRENT_IMAGE(DD, u);
    v = GET_CURRENT_IMAGE(DD, v);

    grad_E_u = GET_CURRENT_IMAGE(DD, grad_E_u);
    grad_E_v = GET_CURRENT_IMAGE(DD, grad_E_v);

    /* For the coarsets level we just iterate */

    if (res == DD->param.max_level)
    {
      printMsg("\n");

      for(i = 0; i <= res; i++)
	printMsg(" ");

      printMsg("Iterating in the level %d ...", res);
      fflush(stdout);

      /* Iterate until convergence */

      if (DD->param.maxit_coarse_multigrid)
	rc = mgtn_tnc(&lastf[res], u, v, grad_E_u, grad_E_v, DD->param.maxit_coarse_multigrid, relative_res, DD);
    }
    else
    {
      /* Are we in a postmoothing stage ? */

      if ((index > 1) && (res < DD->cycle[index - 1]))
      {
	printMsg("\n");

	for(i = 0; i <= res; i++)
	  printMsg(" ");

	printMsg("Postsmoothing in level %d...", res);
	fflush(stderr);

	/*
	 * Compute the correction step and store it in tmp. The
	 * correction step is computed for the coarser level and
	 * is the difference between the current flow (u,v) and
	 * the flow previous to start the V-cycle (restricted_u,v) 
	 */

	nb_pixels = GET_CURRENT_NROW(DD) * GET_CURRENT_NCOL(DD);

	p_tmp = GET_CURRENT_IMAGE(DD, tmp_u)->gray;

	p_d1  = GET_CURRENT_IMAGE(DD, u)->gray;
	p_d2  = GET_CURRENT_IMAGE(DD, restricted_u)->gray;

	for(i = 0; i < nb_pixels; i++, p_tmp++, p_d1++, p_d2++)
	  *p_tmp = *p_d1 - *p_d2; 

	p_tmp = GET_CURRENT_IMAGE(DD, tmp_v)->gray;

	p_d1  = GET_CURRENT_IMAGE(DD, v)->gray;
	p_d2  = GET_CURRENT_IMAGE(DD, restricted_v)->gray;

	for(i = 0; i < nb_pixels; i++, p_tmp++, p_d1++, p_d2++)
	  *p_tmp = *p_d1 - *p_d2;

	/* Interpolate. We prolonge the correction step to the current
	 * resolution level. */

	mgtn_image_prolongation2(GET_CURRENT_IMAGE(DD,tmp_u), GET_CURRENT_IMAGE(DD,tmp_v), GET_FINER_IMAGE(DD,tmp_u), GET_FINER_IMAGE(DD,tmp_v), DD);

	/* Jump to the finer resolution level, we free not necessary memory */

        mgtn_resources_delete_presmooth(DD);
	mgtn_resources_delete_postsmooth(DD);
	DD->curr_level--;
	mgtn_resources_init_postsmooth(DD); 

        /* Setup border conditions */

	mgtn_set_border_conditions_uv(GET_CURRENT_IMAGE(DD,tmp_u), GET_CURRENT_IMAGE(DD,tmp_v));

	/* Initialize {u,v} and {grad_E_u, grad_E_v} */

	u = GET_CURRENT_IMAGE(DD, u);
	v = GET_CURRENT_IMAGE(DD, v);

	grad_E_u = GET_CURRENT_IMAGE(DD, grad_E_u);
	grad_E_v = GET_CURRENT_IMAGE(DD, grad_E_v);

	/* Perform a line search using the correction step previously computed.
	 * We assume that search direction is stored in (tmp_u,tmp_v) of DD. A
	 * -1 is passed as max iterations. The TNC library will then know that
	 * only a line search has to be performed. */

	rc = mgtn_tnc(&lastf[res], u, v, grad_E_u, grad_E_v, -1, relative_res, DD);

	/* We now perform some more iterations. This step is called postsmoothing. */

	if (DD->param.maxit_postsmoothing[res])
	{
	  rc = mgtn_tnc(&lastf[res], u, v, grad_E_u, grad_E_v, DD->param.maxit_postsmoothing[relative_res], relative_res, DD);

	  switch (rc)
	  {
	    case MGTN_TNC_CONVERGED:
	      printMsg(" -> %s", mgtn_rc_string[rc]);
	      fflush(stdout);

	    case MGTN_TNC_USERABORT:
	      do { index++; } while (res <= DD->cycle[index]);
	      index--;
	      break;

	    case MGTN_TNC_MAXITER:
	      break;

	    default:
	      error("postsmoothing: should not arrive here\n");
	  }
	}
	else
	  rc = MGTN_TNC_MAXITER;
      }

      /* Are we in presmoothing ? */

      if (res < DD->cycle[index + 1])
      {
	printMsg("\n");

	for(i = 0; i <= res; i++)
	  printMsg(" ");

	printMsg("Presmoothing in level %d...", res);
	fflush(stdout);

	/* Presmoothing. On output, we store in grad_E_{u,v} the
	 * gradient evaluated at the last point... */

	DD->in_presmoothing = 1;
	rc = mgtn_tnc(&lastf[res], u, v, grad_E_u, grad_E_v, DD->param.maxit_presmoothing[relative_res], relative_res, DD);
	DD->in_presmoothing = 0;

	printMsg(" -> %s", mgtn_rc_string[rc]);
	fflush(stdout);

	switch (rc)
	{
	  case MGTN_TNC_CONVERGED:
	  case MGTN_TNC_USERABORT:

	    /* If we have converged for the current resolutin level, we won't
	     * go down in the resolution level to improve result. There is no
	     * need to do so. */

	    do { index++; } while (res <= DD->cycle[index]);
	    index--;
	    break;

	  case MGTN_TNC_MAXITER:
	  case MGTN_TNC_PRECOND:

	    /* If we have arrived to the maximum number of iterations or the
	     * preconditionning condition is satisfied, there is still work to
	     * do. We go down in the resolution level to obtain a correction
	     * step.
	     */

	    /* Initialize coarser resolution level and deletes current one */

	    mgtn_resources_delete_presmooth(DD); 
	    DD->curr_level++;
	    mgtn_resources_init_presmooth(DD);

	    /* Initialize {u,v} and {grad_E_u, grad_E_v} */

	    res = DD->curr_level;

	    u = GET_CURRENT_IMAGE(DD, u);
	    v = GET_CURRENT_IMAGE(DD, v);

	    grad_E_u = GET_CURRENT_IMAGE(DD, grad_E_u);
	    grad_E_v = GET_CURRENT_IMAGE(DD, grad_E_v);

	    /* Do the restriction of {u,v}, gradient and residue. For the
	     * finest level (in the current V or W-cycle) the residue is
	     * initialized to zero. */

	    mgtn_image_restriction2(GET_FINER_IMAGE(DD,u), GET_FINER_IMAGE(DD,v), GET_CURRENT_IMAGE(DD,restricted_u), GET_CURRENT_IMAGE(DD,restricted_v), DD);
	    mgtn_image_restriction2(GET_FINER_IMAGE(DD,grad_E_u), GET_FINER_IMAGE(DD,grad_E_v), GET_CURRENT_IMAGE(DD,tmp_u), GET_CURRENT_IMAGE(DD,tmp_v), DD);
	    mgtn_image_restriction2(GET_FINER_IMAGE(DD,residue_u), GET_FINER_IMAGE(DD,residue_v), GET_CURRENT_IMAGE(DD,residue_u), GET_CURRENT_IMAGE(DD,residue_v), DD); 

	    /* Set borders correctly */

	    mgtn_set_border_conditions_uv(GET_CURRENT_IMAGE(DD,restricted_u), GET_CURRENT_IMAGE(DD,restricted_v)); 

	    mgtn_set_border_conditions_gradient(GET_CURRENT_IMAGE(DD,residue_u), GET_CURRENT_IMAGE(DD,residue_v));
	    mgtn_set_border_conditions_gradient(GET_CURRENT_IMAGE(DD,tmp_u), GET_CURRENT_IMAGE(DD,tmp_v));

	    /* Initialize motion field of current level using restriction */

	    mgtn_image_copy(GET_CURRENT_IMAGE(DD,restricted_u), u);
	    mgtn_image_copy(GET_CURRENT_IMAGE(DD,restricted_v), v);

	    /* Compute the gradient of the objective function at {u,v}
	     * at the current (coarser) level. */

	    lastf[res] = mgtn_functional_evaluation(u, v, DD);
	    mgtn_gradient_evaluation(u, v, GET_CURRENT_IMAGE(DD, grad_E_u), GET_CURRENT_IMAGE(DD, grad_E_v), DD);

	    DD->nfeval[res]++;
	    DD->ngeval[res]++;

	    /* Compute residual */

	    nb_pixels = GET_CURRENT_NROW(DD) * GET_CURRENT_NCOL(DD);

	    p_d1 = GET_CURRENT_IMAGE(DD, tmp_u)->gray;
	    p_d2 = grad_E_u->gray;
	    p_d3 = GET_CURRENT_IMAGE(DD, residue_u)->gray;

	    for(i = 0; i < nb_pixels; i++, p_d1++, p_d2++, p_d3++)
	      *p_d3 = *p_d2 - *p_d1;

	    p_d1 = GET_CURRENT_IMAGE(DD, tmp_v)->gray;
	    p_d2 = grad_E_v->gray;
	    p_d3 = GET_CURRENT_IMAGE(DD, residue_v)->gray;

	    for(i = 0; i < nb_pixels; i++, p_d1++, p_d2++, p_d3++)
	      *p_d3 = *p_d2 - *p_d1;

	    /* Set border conditions */

	    mgtn_set_border_conditions_gradient(GET_CURRENT_IMAGE(DD,residue_u), GET_CURRENT_IMAGE(DD,residue_v));

	    /* Compute functional value */

	    lastf[res] = mgtn_functional_evaluation_multigrid(u, v, GET_CURRENT_IMAGE(DD, residue_u), GET_CURRENT_IMAGE(DD, residue_v), DD);
	    mgtn_gradient_evaluation_multigrid(u, v, GET_CURRENT_IMAGE(DD, residue_u), GET_CURRENT_IMAGE(DD, residue_v), grad_E_u, grad_E_v, DD);

	    break;

	  default:
	    error("presmoothing: should not arrive here\n");
	}
      }
    }

    index++;

  } /* end while */

  free(lastf);

  return rc;
}

