//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 * 
 * Set input images 
 *
 */

void mgtn_set_input_images(
  struct image *in0,
  struct image *in1,
  struct resources *DD)
{
  DD->ori_in0 = in0;
  DD->ori_in1 = in1;
}

/**
 * 
 * Allocation and initialization of the resources structure
 *
 */

void mgtn_resources_initialize(
  struct resources *DD)
{
  int i, nelems;

  /* Initialization of fields to zero or null */  

  DD->curr_level       = 0;
  DD->curr_level_cycle = 0;

  DD->t_active = 0;
  DD->t_elapsed = 0;
  DD->t_last = 0;

  DD->nfeval = NULL;
  DD->ngeval = NULL;

  DD->in_multigrid = 0;

  DD->nrow = NULL;
  DD->ncol = NULL;

  DD->in0 = NULL; 
  DD->in1 = NULL; 

  DD->grad_t0_x  = NULL; 
  DD->grad_t0_y  = NULL; 
  DD->grad_t1_x  = NULL; 
  DD->grad_t1_y  = NULL;
  DD->grad_t1_xx = NULL;
  DD->grad_t1_yy = NULL;
  DD->grad_t1_xy = NULL;

  DD->u = NULL;
  DD->v = NULL;

  DD->have_groundtruth_flow = 0;
  DD->groundtruth_flow_u = NULL;
  DD->groundtruth_flow_v = NULL;

  DD->grad_E_u = NULL;
  DD->grad_E_v = NULL;

  DD->residue_u = NULL;
  DD->residue_v = NULL;

  DD->tmp_u = NULL;
  DD->tmp_v = NULL;

  DD->restricted_u = NULL;
  DD->restricted_v = NULL;

  /* Check if multigrid is going to be used */ 

  switch (DD->param.scheme)
  {
    case SCHEME_MULTIRESOLUTION:
      DD->in_multigrid = 0;
      break;

    default:
      DD->in_multigrid = 1;
  }

  nelems = DD->param.max_level + 1;

  /* Allocate memory for nfeval, ngeval */

  DD->nfeval = (int *) xmalloc(sizeof(int) * nelems);
  DD->ngeval = (int *) xmalloc(sizeof(int) * nelems);

  /* Allocate memory for nrow, ncol */

  DD->nrow = (int *) xmalloc(sizeof(int) * nelems);
  DD->ncol = (int *) xmalloc(sizeof(int) * nelems);

  /* Allocate memory for multiresolution images */

  DD->in0  = (struct image **) xmalloc(sizeof(struct image *) * nelems);
  DD->in1  = (struct image **) xmalloc(sizeof(struct image *) * nelems);

  /* Allocate memory for u, v */

  DD->u = (struct image **) xmalloc(sizeof(struct image *) * nelems);
  DD->v = (struct image **) xmalloc(sizeof(struct image *) * nelems);

  /* Allocate memory for gradient images */

  DD->grad_t0_x = (struct image **) xmalloc(sizeof(struct image *) * nelems);
  DD->grad_t0_y = (struct image **) xmalloc(sizeof(struct image *) * nelems);

  DD->grad_t1_x = (struct image **) xmalloc(sizeof(struct image *) * nelems);
  DD->grad_t1_y = (struct image **) xmalloc(sizeof(struct image *) * nelems);

  DD->grad_t1_xx = (struct image **) xmalloc(sizeof(struct image *) * nelems);
  DD->grad_t1_yy = (struct image **) xmalloc(sizeof(struct image *) * nelems);
  DD->grad_t1_xy = (struct image **) xmalloc(sizeof(struct image *) * nelems);

  /* Allocate memory for grad_E */

  DD->grad_E_u = (struct image **) xmalloc(sizeof(struct image *) * nelems);
  DD->grad_E_v = (struct image **) xmalloc(sizeof(struct image *) * nelems);

  if (DD->in_multigrid)
  {
    /* Allocate memory for the residue */

    DD->residue_u = (struct image **) xmalloc(sizeof(struct image *) * nelems);
    DD->residue_v = (struct image **) xmalloc(sizeof(struct image *) * nelems);

    /* Allocate memory for temporal data */

    DD->tmp_u = (struct image **) xmalloc(sizeof(struct image *) * nelems);
    DD->tmp_v = (struct image **) xmalloc(sizeof(struct image *) * nelems);

    /* Allocate memory for temporal data */

    DD->restricted_u = (struct image **) xmalloc(sizeof(struct image *) * nelems);
    DD->restricted_v = (struct image **) xmalloc(sizeof(struct image *) * nelems);
  }

  /* Initialize images and remaining members */

  for(i = DD->param.max_level; i >= 0; i--)
  {
    DD->nfeval[i] = 0;
    DD->ngeval[i] = 0;

    DD->nrow[i] = 0;
    DD->ncol[i] = 0;

    DD->in0[i] = NULL;
    DD->in1[i] = NULL;

    DD->u[i] = NULL;
    DD->v[i] = NULL;

    DD->grad_E_u[i] = NULL;
    DD->grad_E_v[i] = NULL;

    DD->grad_t0_x[i] = NULL;
    DD->grad_t0_y[i] = NULL;
    DD->grad_t1_x[i] = NULL;
    DD->grad_t1_y[i] = NULL;

    DD->grad_t1_xx[i] = NULL; 
    DD->grad_t1_yy[i] = NULL; 
    DD->grad_t1_xy[i] = NULL; 

    if (DD->in_multigrid)
    {
      DD->residue_u[i] = NULL;
      DD->residue_v[i] = NULL;

      DD->tmp_u[i] = NULL;
      DD->tmp_v[i] = NULL;

      DD->restricted_u[i] = NULL;
      DD->restricted_v[i] = NULL;
    }
  }

  /* Initialize nrow and ncol */

  mgtn_initialize_nrow_ncol(DD->ori_in0->nrow, DD->ori_in0->ncol, DD);

  /* Initialize temporal images for use in multigrid only */

  if (DD->in_multigrid)
  {
    for(i = DD->param.max_level; i >= 0; i--)
    {
      DD->tmp_u[i] = mgtn_image_alloc(NULL, DD->nrow[i], DD->ncol[i]);
      DD->tmp_v[i] = mgtn_image_alloc(NULL, DD->nrow[i], DD->ncol[i]);
    }
  }
}

/**
 *
 * Loads into memory the data for performing the premooth step at
 * a given resolution level.  For the case of multiresolution
 * calling this function allows to load the necessary data to
 * perform interations at a given resolution level.
 *
 */

void mgtn_resources_init_presmooth(
    struct resources * DD)
{
  int res, nrow, ncol;
  clock_t t2;

  /* This is a workaround to avoid counting the time
   * needed to load the data into memory.
   */

  t2 = 0;

  if (DD->t_active)
  {
    t2 = clock();
    DD->t_elapsed += t2 - DD->t_last; 
  }

  res = DD->curr_level;

  nrow = GET_CURRENT_NROW(DD);
  ncol = GET_CURRENT_NCOL(DD);

  DD->u[res] = mgtn_image_alloc(NULL, nrow, ncol);
  DD->v[res] = mgtn_image_alloc(NULL, nrow, ncol);

  DD->grad_E_u[res] = mgtn_image_alloc(NULL, nrow, ncol);
  DD->grad_E_v[res] = mgtn_image_alloc(NULL, nrow, ncol);

  mgtn_image_clear(DD->grad_E_u[res], 0.0);
  mgtn_image_clear(DD->grad_E_v[res], 0.0);

#ifdef LOW_MEMORY_COMPUTER
  mgtn_read_gradient_images(DD);
#endif

  if (DD->in_multigrid)
  {
    DD->residue_u[res] = mgtn_image_alloc(NULL, nrow, ncol);
    DD->residue_v[res] = mgtn_image_alloc(NULL, nrow, ncol);

    if (res != 0)
    {
      DD->restricted_u[res] = mgtn_image_alloc(NULL, nrow, ncol);
      DD->restricted_v[res] = mgtn_image_alloc(NULL, nrow, ncol);
    }

    /* Clear data */

    mgtn_image_clear(DD->residue_u[res], 0.0);
    mgtn_image_clear(DD->residue_v[res], 0.0);
  }

  if (DD->t_active)
    DD->t_last = clock();
}

/**
 *
 * Loads into memory the data for performing the postsmoothing
 * step at a given resolution level. Only for multigrid. 
 *
 */

void mgtn_resources_init_postsmooth(
    struct resources *DD)
{
  clock_t t2;

  /* This is a workaround to avoid counting the time
   * needed to load the data into memory.
   */

  t2 = 0;

  if (DD->t_active)
  {
    t2 = clock();
    DD->t_elapsed += t2 - DD->t_last; 
  }

  /* Read gradient images */

#ifdef LOW_MEMORY_COMPUTER
  mgtn_read_gradient_images(DD);
#endif

  if (DD->t_active)
    DD->t_last = clock();
}

/**
 *
 *  Deletes from memory some data after finishing the presmoothing
 *  step. Also used to delete data for classical multiresolution
 *  when going from one step to the next.
 *
 */

void mgtn_resources_delete_presmooth(
    struct resources *DD)
{
  int res;
  clock_t t2;

  /* This is a workaround to avoid counting the time
   * needed to delete the data from memory 
   */

  t2 = 0;

  if (DD->t_active)
  {
    t2 = clock();
    DD->t_elapsed += t2 - DD->t_last; 
  }

  res = DD->curr_level;

#ifdef LOW_MEMORY_COMPUTER
  mgtn_gradient_images_delete(DD);
#endif

  if (DD->t_active)
    DD->t_last = clock();
}

/**
 *
 *  Deletes from memory some data after finishing the
 *  postsmoothing step. This function deletes all the data
 *  necessary for doing computations at a certain level. This
 *  function also is called by multiresolution algorithm when
 *  finishing a certain level.
 *
 */

void mgtn_resources_delete_postsmooth(
    struct resources * DD)
{
  int res;
  clock_t t2;

  /* This is a workaround to avoid counting the time
   * needed to delete the data from memory 
   */

  t2 = 0;

  if (DD->t_active)
  {
    t2 = clock();
    DD->t_elapsed += t2 - DD->t_last; 
  }

  res = DD->curr_level;

  if (DD->u[res])
  {
    mgtn_image_delete(DD->u[res]);
    mgtn_image_delete(DD->v[res]);

    DD->u[res] = NULL;
    DD->v[res] = NULL;
  }

  if (DD->grad_E_u[res])
  {
    mgtn_image_delete(DD->grad_E_u[res]);
    mgtn_image_delete(DD->grad_E_v[res]);

    DD->grad_E_u[res] = NULL;
    DD->grad_E_v[res] = NULL;
  }

  if (DD->in_multigrid)
  {
    if (DD->residue_u[res])
    {
      mgtn_image_delete(DD->residue_u[res]);
      mgtn_image_delete(DD->residue_v[res]);

      DD->residue_u[res] = NULL; 
      DD->residue_v[res] = NULL;
    }

    if ((res != 0) && (DD->restricted_u[res]))
    {
      mgtn_image_delete(DD->restricted_u[res]);
      mgtn_image_delete(DD->restricted_v[res]);

      DD->restricted_u[res] = NULL; 
      DD->restricted_v[res] = NULL;
    }
  }

  if (DD->t_active)
    DD->t_last = clock();
}

/**
 *
 *  Deletes the resources. 
 *
 */

void mgtn_resources_delete(
    struct resources *DD)
{
  int i;
#ifdef LOW_MEMORY_COMPUTER
  char fname[200];
#endif

  /* Remove remaining data */ 

  for(i = DD->param.max_level; i >= 0; i--)
  {
    DD->curr_level = i;

    mgtn_resources_delete_presmooth(DD);
    mgtn_resources_delete_postsmooth(DD); 

    mgtn_image_delete(DD->in0[i]);
    mgtn_image_delete(DD->in1[i]);

    if (DD->in_multigrid)
    {
      mgtn_image_delete(DD->tmp_u[i]);
      mgtn_image_delete(DD->tmp_v[i]);

    }

    /* Remove gradient image */

#ifdef LOW_MEMORY_COMPUTER
    sprintf(fname, "gradient_%d.data", i);
    unlink(fname);
#endif

   /* Free gradient image */
#ifndef LOW_MEMORY_COMPUTER
    mgtn_gradient_images_delete(DD);
#endif
  }

  if (DD->in_multigrid) {
    free(DD->param.maxit_v_cycle);
    free(DD->param.maxit_presmoothing);
    free(DD->param.maxit_postsmoothing); 

    free(DD->cycle);
  }

  if (DD->groundtruth_flow_u)
  {
    mgtn_image_delete(DD->groundtruth_flow_u);
    mgtn_image_delete(DD->groundtruth_flow_v);
  }

  free(DD->nfeval);
  free(DD->ngeval);

  free(DD->nrow);
  free(DD->ncol);

  free(DD->u);
  free(DD->v);

  free(DD->in0);
  free(DD->in1);

  free(DD->grad_E_u);
  free(DD->grad_E_v);

  free(DD->grad_t0_x);
  free(DD->grad_t0_y);
  free(DD->grad_t1_x);
  free(DD->grad_t1_y);

  free(DD->grad_t1_xx);
  free(DD->grad_t1_yy);
  free(DD->grad_t1_xy);

  if (DD->in_multigrid)
  {
    free(DD->residue_u);
    free(DD->residue_v);

    free(DD->restricted_u);
    free(DD->restricted_v);

    free(DD->tmp_u);
    free(DD->tmp_v);
  }
}

