//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"
#include "tnc.h"

/**
 *
 * The functions in this file provide the interface to the TNC
 * library.
 *
 */


/**
 *
 * This is the interface function between our algorithm (controlled by
 * the main function) and the Truncated Newton algorithm (see file tnc.c).
 * The function prepares the data to be processed by the TNC library.
 *
 */

int mgtn_tnc(
    double *f, 
    struct image *u, 
    struct image *v, 
    struct image *g_u, 
    struct image *g_v, 
    int maxiter, 
    int relative_res,
    struct resources *DD)
{
  struct image *residue_u, *residue_v;

  double beta;
  double *low, *up, *x, *g, *p_double;

  int rc, nfeval, maxnfeval = MAXNFEVAL;
  double fmin, eta = -1.0, stepmx = -1.0, accuracy = -1.0, 
	 rescale = -1.0;

  int i, nb_pixels, nb_pixels2, nrow, ncol;
  int min_x, min_y, max_x, max_y;

  /* Initialization of stopping parameters */

  double ftol  = DD->param.ftol;
  double xtol  = DD->param.xtol;
  double pgtol = DD->param.pgtol;

  int maxCGit  = DD->param.maxit_tn_internal;

  /* Initiate residue in case we are in multigrid */

  if (DD->in_multigrid)
  {
    residue_u = GET_CURRENT_IMAGE(DD, residue_u);
    residue_v = GET_CURRENT_IMAGE(DD, residue_v);
  }
  else
  {
    residue_u = NULL;
    residue_v = NULL;
  }

  /* In order to pass the data to the TNC library we "discard" the
   * pixels at the border of the image since these pixels are used
   * for border conditions.
   */

  nrow = GET_CURRENT_NROW(DD);
  ncol = GET_CURRENT_NCOL(DD);

  min_x = 1;
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  nb_pixels  = (max_x - min_x + 1) * (max_y - min_y + 1);
  nb_pixels2 = nb_pixels * 2;

  /* Allocate memory for the vectors the truncated Newton method needs */

  x = (double *) xmalloc(sizeof(double) * nb_pixels2);
  g = (double *) xmalloc(sizeof(double) * nb_pixels2);

  low = (double *) xmalloc(sizeof(double) * nb_pixels2);
  up  = (double *) xmalloc(sizeof(double) * nb_pixels2);

  if (DD->in_multigrid)
    beta = DD->param.beta_constraint;
  else
    beta = -1.0;

  if ((relative_res == 0) || (beta < 0.0))
  {
    p_double = up;
    for(i = 0; i < nb_pixels2; i++, p_double++)
      *p_double = HUGE_VAL; 

    p_double = low;
    for(i = 0; i < nb_pixels2; i++, p_double++)
      *p_double = - HUGE_VAL;

    fmin = 0.0;
  }
  else
  {
    /* Compute upper and lower bound. Note that bound are computed over the
     * restricted field, not the current iteration. */

    mgtn_merge_vector_uv(GET_CURRENT_IMAGE(DD,restricted_u), GET_CURRENT_IMAGE(DD,restricted_v), up);
    memcpy(low, up, sizeof(double) * nb_pixels2);

    p_double = up;
    for(i = 0; i < nb_pixels2; i++, p_double++)
      *p_double += beta; 

    p_double = low;
    for(i = 0; i < nb_pixels2; i++, p_double++)
      *p_double -= beta;

    /* Compute fmin, minimum possible value for the energy. We assume that
     * the objective energy can reach the minimal 0 value. */

    fmin = 0.0;

    p_double = residue_u->gray;
    for(i = 0; i < nb_pixels; i++, p_double++)
      fmin -= MAX(*p_double * up[i], *p_double * low[i]);

    p_double = residue_v->gray;
    for(i = nb_pixels; i < nb_pixels2; i++, p_double++)
      fmin -= MAX(*p_double * up[i], *p_double * low[i]);
  }

  /* We prepare data for the TNC library. {u,v} and {g_u, g_v} components are
   * merged in one vector
   */

  mgtn_merge_vector_uv(u, v, x);
  mgtn_set_border_conditions_uv(u, v);

  mgtn_merge_vector_uv(g_u, g_v, g);
  mgtn_set_border_conditions_gradient(u, v);

  /* Print current value of function to screen */

  printMsg(" f_in = %E", *f);
  fflush(stdout);

  /* Call TNC library */

  rc = tnc(nb_pixels2, x, f, g, mgtn_tnc_function, (void *) DD, low, up, 
      NULL, NULL, TNC_MSG_NONE, maxiter, maxCGit, maxnfeval, eta, stepmx, 
      accuracy, fmin, ftol, xtol, pgtol, rescale, &nfeval);

  /* Split data that comes from TNC library to our format */

  mgtn_split_vector_uv(x, u, v);
  mgtn_split_vector_uv(g, g_u, g_v);

  /* Print output */

  if (maxiter < 0)
    printMsg(" f_lin_min = %5E", *f);
  else
    printMsg(" f_out = %E", *f);

  fflush(stdout);

  free(x);
  free(g);
  free(low);
  free(up);

  switch (rc)
  {
    case TNC_FCONVERGED:
    case TNC_XCONVERGED:
    case TNC_LOCALMINIMUM:
      i = MGTN_TNC_CONVERGED;
      break;

    case TNC_USERABORT:
      i = MGTN_TNC_USERABORT;
      break;

    case TNC_MAXFUN:
      i = MGTN_TNC_MAXITER;
      break;

    case TNC_PRECOND:
      i = MGTN_TNC_PRECOND;
      break;

    case TNC_CORRNOTDESC:
      i = MGTN_TNC_CORRNOTDESC;
      break;

    case TNC_NOPROGRESS:
      error("\nTNC could not progress. Sorry.\n");

    default:
      fprintf(stderr, "\nTNC returned the following error code which could not be handled: %d\n", rc);
      error("Cannot progress. Sorry.\n");
  }

  return i;
}

/**
 *
 *  This function is called by the TNC library in order to evaluate the functional
 *  and the gradient of the energy.
 *
 */

int mgtn_tnc_function(
    double *x, 
    double *f, 
    double *g, 
    void *state)
{
  struct resources *DD;
  struct image *u, *v, *grad_E_u, *grad_E_v;

  /* Get DD (resources) thanks to 'state' parameter */

  DD = (struct resources *) state;

  u        = GET_CURRENT_IMAGE(DD,u);
  v        = GET_CURRENT_IMAGE(DD,v);
  grad_E_u = GET_CURRENT_IMAGE(DD,grad_E_u);
  grad_E_v = GET_CURRENT_IMAGE(DD,grad_E_v);

  mgtn_split_vector_uv(x, u, v);

  mgtn_set_border_conditions_uv(u, v);

  if (DD->in_multigrid)
  {
    *f = mgtn_functional_evaluation_multigrid(u, v, GET_CURRENT_IMAGE(DD, residue_u), GET_CURRENT_IMAGE(DD, residue_v), DD);
    mgtn_gradient_evaluation_multigrid(u, v, GET_CURRENT_IMAGE(DD, residue_u), GET_CURRENT_IMAGE(DD, residue_v), grad_E_u, grad_E_v, DD);
  }
  else
  {
    *f = mgtn_functional_evaluation(u, v, DD);
    mgtn_gradient_evaluation(u, v, grad_E_u, grad_E_v, DD);
  }

  mgtn_merge_vector_uv(grad_E_u, grad_E_v, g);

  return 0;
}

/**
 *
 *  Increments the number of function evaluations. This function is called by
 *  the TNC library when TNC evaluates the energy functional.
 *
 */

void mgtn_tnc_nfeval(
    void *state)
{
  struct resources *DD;

  DD = (struct resources *) state;

  DD->nfeval[DD->curr_level]++;
}

/**
 *
 *  Increments the number of gradient evaluations. This function is called by
 *  the TNC library when TNC evaluates the gradient of the functional.
 *
 */

void mgtn_tnc_ngeval(state)
  void *state;
{
  struct resources *DD;

  DD = (struct resources *) state;

  DD->ngeval[DD->curr_level]++;
}

/**
 *
 *  Used only for multigrid. This function is called by TNC library to check if
 *  the preconditioning conditions is satisfied. The preconditioning conditions
 *  allow to automatically decide if the algorithm has performed enough
 *  presmmothing interations in order to restrict and continue iterations at a
 *  lower resolution level.
 *
 */

int mgtn_tnc_chkprecond(
    double *g,
    void *state)
{
  struct resources *DD;
  struct image *tmp_u, *tmp_v, *tmp_coarser_u, *tmp_coarser_v;

  double g_norm, g_restricted_norm, aaa, eps;

  DD = (struct resources *) state;

  if (!DD->in_presmoothing) 
    return 0;

  if ((DD->param.e_presmooth_cond < 0.0) || (DD->param.a_presmooth_cond < 0.0))
    return 0;

  tmp_u = GET_CURRENT_IMAGE(DD,tmp_u);
  tmp_v = GET_CURRENT_IMAGE(DD,tmp_v);

  tmp_coarser_u = GET_COARSER_IMAGE(DD,tmp_u);
  tmp_coarser_v = GET_COARSER_IMAGE(DD,tmp_v);

  mgtn_split_vector_uv(g, tmp_u, tmp_v);
  mgtn_set_border_conditions_gradient(tmp_u, tmp_v);

  mgtn_image_restriction2(tmp_u, tmp_v, tmp_coarser_u, tmp_coarser_v, DD);
  mgtn_set_border_conditions_gradient(tmp_coarser_u, tmp_coarser_v);

  g_norm = mgtn_gradient_norm(tmp_u, tmp_v);
  g_restricted_norm = mgtn_gradient_norm(tmp_coarser_u, tmp_coarser_v);

  aaa = DD->param.a_presmooth_cond / SQR(DD->curr_level+1);
  eps = DD->param.e_presmooth_cond / SQR(DD->curr_level+1);

  if (g_restricted_norm < eps)
    return 0;

  if (g_restricted_norm < aaa * g_norm)
    return 0;

  return 1;
}

/**
 *
 *  Used only for multigrid. Used for the postsmoothing iterations, in which
 *  first a line search is performed. The search direction given by the line
 *  search comes from the computations performed in the coarser level. Here we
 *  have implemented this in a not so nice way.  The search direction is stored
 *  in tmp_u and tmp_v, and has been computed previously in the function
 *  mgtn_multigrid.  Once in the TNC, a parameter tells the TNC library that
 *  only a line search has to be performed and that this line search is given
 *  by the vector stored in tmp_u, tmp_v.  TNC then calls this function to
 *  recover the line search direction.
 *
 */

int mgtn_tnc_get_linsrchdir(p, state)
  double *p;
  void *state;
{
  struct resources *DD;

  DD = (struct resources *) state;
  mgtn_merge_vector_uv(GET_CURRENT_IMAGE(DD,tmp_u), GET_CURRENT_IMAGE(DD,tmp_v), p);

  return 0;
}

/**
 *
 *  Used only in multigrid. The function mgtn_tnc_get_linsrchdir allows to
 *  recover the line search direction to be used during the postsmmothing.
 *  The parameter check_descent (see code) can be configured through the
 *  parameter file and allows to make postsmoothing faster. If this parameter
 *  is true, the algorithm checks if the search direction (with a step = 1)
 *  already descreases the energy value. If so, the update is accepted.
 *  Otherwise, the algorithm performs a usual line search by trying to find
 *  a good step value. If this parameter is false, then the TNC algorithm
 *  will always try to find a good value for step even if for step = 1
 *  the energy decreases. 
 *
 */

int mgtn_tnc_chkdescent(
    void *state)
{
  int rc;
  struct resources *DD;

  DD = (struct resources *) state;
  rc = (DD->param.check_descent == MG_CHECK_DESCENT_YES);

  return rc;
}


