//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

extern int quiet;

#ifndef M_PI
#define M_PI		3.14159265358979323846	/* pi */
#endif

/**
 *
 * This function is like "malloc", but it returns always a valid pointer
 * Code by Enric Meinhardt.
 *
 */

void *xmalloc(size_t size)
{
  void *p = malloc(size);
  if (!p) 
    error("xmalloc: out of memory\n");

  return p;
}

/**
 * 
 * Faltal error, print a message to stderr and exit
 *
 */

void error(char *msg)
{
  fprintf(stderr,"%s\n",msg);
  exit(EXIT_FAILURE);
}

/**
 *
 * Print message function. Message is printed on screen depending
 * on value of 'quiet', a global variable.
 *
 */

void printMsg( const char *pFormat, ... )
{
  va_list args;

  if (!quiet)
  {
    va_start(args, pFormat);
    (void)vprintf(pFormat, args);
    va_end(args);
  }
}

/**
 *
 *  This function is used to initialize the nrow and ncol vectors,
 *  that is, the size (row and columns) of the images along the 
 *  pyramid.
 *
 */

void mgtn_initialize_nrow_ncol(
    int nrow, 
    int ncol, 
    struct resources *DD)
{
  int i, new_nrow, new_ncol;

  DD->nrow[0] = nrow;
  DD->ncol[0] = ncol;

  for(i = 1; i <= DD->param.max_level; i++)
  {
    if (EVEN(ncol))
      new_ncol = ncol >> 1;
    else
      new_ncol = (ncol + 1) >> 1;

    if (EVEN(nrow))
      new_nrow = nrow >> 1;
    else
      new_nrow = (nrow + 1) >> 1;

    DD->nrow[i] = new_nrow;
    DD->ncol[i] = new_ncol;

    nrow = new_nrow;
    ncol = new_ncol;
  }
}

/**
 *
 *  This function is used to copy the flow from the format used by the main
 *  function to the truncated Newton function.  Indeed, the truncated Newton
 *  function needs flow (u and v) to be merged in one vector. Only inner pixels
 *  are copied since border pixels are used for border conditions.
 *
 */

void mgtn_split_vector_uv(
    double *x, 
    struct image *u, 
    struct image *v)
{
  double *p_base, *p_uv;
  int i, j, nrow, ncol, offset;
  int min_x, min_y, max_x, max_y;

  nrow = u->nrow;
  ncol = u->ncol;

  min_x = 1;
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  p_base = u->gray;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;
    p_uv   = p_base + offset;

    for(j = min_x; j <= max_x; j++, p_uv++, x++)
      *p_uv = *x;
  }

  p_base = v->gray;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;
    p_uv   = p_base + offset;

    for(j = min_x; j <= max_x; j++, p_uv++, x++)
      *p_uv = *x;
  }
}

/**
 *
 *  This function is used to copy the flow from the format used by the
 *  truncated Newton function to the one used by the main function.  Indeed,
 *  the truncated Newton function needs flow (u and v) to be merged in one
 *  vector. Only inner pixels are copied since border pixels are used for
 *  border conditions.
 *
 */

void mgtn_merge_vector_uv(
    struct image *u, 
    struct image *v, 
    double *x)
{
  double *p_base, *p_uv;
  int i, j, nrow, ncol, offset;
  int min_x, min_y, max_x, max_y;

  nrow = u->nrow;
  ncol = u->ncol;

  min_x = 1;
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  p_base = u->gray;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;
    p_uv   = p_base + offset;

    for(j = min_x; j <= max_x; j++, p_uv++, x++)
      *x = *p_uv;
  }

  p_base = v->gray;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;
    p_uv   = p_base + offset;

    for(j = min_x; j <= max_x; j++, p_uv++, x++)
      *x = *p_uv;
  }
}

/**
 *
 *  Initialize flow at current level to zero
 *
 */

void mgtn_flow_initialize(
    struct resources *DD)
{
  struct image *u, *v;
  struct vector2 coord_pix;

  int nrow, ncol;
  double *p_u,*p_v;

  u = GET_CURRENT_IMAGE(DD,u);
  v = GET_CURRENT_IMAGE(DD,v);

  p_u = u->gray;
  p_v = v->gray;

  nrow = u->nrow;
  ncol = u->ncol;

  for(coord_pix.y = 0; coord_pix.y < nrow; coord_pix.y++)
    for(coord_pix.x = 0; coord_pix.x < ncol; coord_pix.x++, p_u++, p_v++)
    {
      *p_u = 0.0; 
      *p_v = 0.0;
    }
}


/**
 * 
 * This function returns 1 if argument is a power
 * of two.
 *
 */

int mgtn_fourier_power_of_two(
    int n)
{
  if (n<1) return(0);
  while ((n&1)==0) n=(n>>1);
  return(n==1);
}

/**
 *
 * Compute the FFT of a data array. This function is used to
 * interpolate the filter taps (for computing the gradient) at non
 * integer points. Output vectors have to be allocated.
 *
 */

#define SWAP(a,b,tmp)        	 (tmp)=(a) ; (a)=(b) ; (b)=(tmp); 

void mgtn_fourier_fft1d(
    struct vector *Xr,
    struct vector *Xi,
    struct vector *Yr,
    struct vector *Yi,
    int inverse)
{
  int size,n,mmax,m,j,istep,i,isign;
  double wtemp,wr,wpr,wpi,wi,theta;
  float tempr,tempi;
  double *data;

  size = Xr->size;

  if (!mgtn_fourier_power_of_two(size)) 
    error("Size of input signal must be a power of 2 !\n");

  if (Xi) if (Xi->size!=size) 
    error("Real and imaginary parts are of different size !\n");

  if (!inverse) isign=1; else isign=-1;

  data = (double *) malloc(2*size*sizeof(double));

  for (i=0;i<size;i++)
  {
    data[2*i]   = Xr->values[i];
    data[2*i+1] = (Xi?Xi->values[i]:0.0);
  }

  /*--- compute FFT of "data" array ---*/

  n=size << 1;
  j=1;
  for (i=1;i<n;i+=2) {
    if (j>i) {
      SWAP(data[j-1],data[i-1], tempr);
      SWAP(data[j],data[i], tempr);
    }
    m=n >> 1;
    while (m >=2 && j>m) {
      j=j-m;
      m >>=1;
    }
    j=j+m;
  }
  mmax=2;
  while (n>mmax) {
    istep=2*mmax;
    theta=2*M_PI/(isign*mmax);
    wtemp=sin(0.5*theta);
    wpr=-2.0*wtemp*wtemp;
    wpi=sin(theta);
    wr=1.0;
    wi=0.0;
    for (m=1;m<mmax;m+=2) {
      for (i=m-1;i<=n-1;i+=istep) {
	j=i+mmax;
	tempr=wr*data[j]-wi*data[j+1];
	tempi=wr*data[j+1]+wi*data[j];
	data[j]=data[i]-tempr;
	data[j+1]=data[i+1]-tempi;
	data[i]=data[i]+tempr;
	data[i+1]=data[i+1]+tempi;
      }
      wr=(wtemp=wr)*wpr-wi*wpi+wr;
      wi=wi*wpr+wtemp*wpi+wi;
    }
    mmax=istep;
  }

  /*--- store result ---*/
  if (Yr) {
    if (inverse) 
      for (i=0;i<size;i++)
	Yr->values[i] = (float) (data[2*i] / (double) size);
    else
      for (i=0;i<size;i++)
	Yr->values[i]= (float) (data[2*i]);
  }
  if (Yi) {
    if (inverse) 
      for (i=0;i<size;i++)
	Yi->values[i]=(float) (data[2*i+1]/(double)size);
    else 
      for (i=0;i<size;i++)
	Yi->values[i]=(float) data[2*i+1];
  }

  free((char *) data);
}

/** 
 *
 * Perform a non-integer shift of filter. The result is stored in
 * filter_shift, which has to be previously allocated. This function
 * is used for computing the filter taps for the gradient.
 *
 */

void mgtn_fourier_nonintegershift(
    struct vector *filter, 
    double shift, 
    struct vector *filter_shift)
{
  int i, k, N;
  double angle, a, b;

  struct vector *freq_filterR, *freq_filterI;

  /* Discrete Fourier Transform of filter */  

  N = filter->size;

  freq_filterR = mgtn_vector_alloc(NULL, N); 
  freq_filterI = mgtn_vector_alloc(NULL, N); 

  mgtn_fourier_fft1d(filter, NULL, freq_filterR, freq_filterI, 0);

  /* Shift in the Fourier domain */

  angle = - 2.0 * M_PI * shift / (double) N;

  k = 0;
  for(i=0; i< N; i++, k++)
  {
    if(i == N/2)
      k=-i;

    a = freq_filterR->values[i] * cos(angle * (double) k) - 
      freq_filterI->values[i] * sin(angle * (double) k);
    b = freq_filterR->values[i] * sin(angle * (double) k) + 
      freq_filterI->values[i] * cos(angle * (double) k);

    freq_filterR->values[i] = (float) a;
    freq_filterI->values[i] = (float) b;
  }

  freq_filterI->values[N/2] = 0.0; 

  /* Inverse transfrom */

  mgtn_fourier_fft1d(freq_filterR, freq_filterI, filter_shift, NULL, 1); 

  mgtn_vector_delete(freq_filterR);
  mgtn_vector_delete(freq_filterI);
}

/**
 *
 *  Get filter coefficients for prefiltering of image. Based on
 *  a simple gauss filter.
 *
 */

struct vector *mgtn_get_filter_prefilter(
    int length,
    double sigma)
{
  int i, x;
  double den, fac, value;
  struct vector *prefilter;

  if (!ODD(length)) 
    error("Filter length for prefilter should be odd.\n");

  prefilter = mgtn_vector_alloc(NULL, length);
  prefilter->shift = - (length - 1) >> 1;

  den = 1.0 / (sqrt(2*M_PI)*sigma);
  fac = 2.0 * sigma * sigma;

  for(i = 0; i < length; i++) {
    x = i + prefilter->shift;
    value = den * exp(- (double) (x * x) / fac);
    prefilter->values[i] = value;
  }

  return prefilter;
}

/**
 * 
 *  Get filter coefficients for gradient computation
 *
 */

struct vector *mgtn_get_filter_gradient(
    int filter_type, 
    int derivative_order, 
    int length, 
    int index, 
    int interpolated_samples)
{
  struct vector *filter, *filter_zero_padded, *filter_shifted;

  int i, j, k, eff_length;
  double shift, *my_filter; 

  /* Simoncelli filter taps, paper "Differentiation of multi-dimensional
   * signals" */

  double simoncelli_first_5taps[2][5] = 
  {{0.037659, 0.249153, 0.426375, 0.249153, 0.037659},
    {-0.109604, -0.276691, 0.0, 0.276691, 0.109604}};

  double simoncelli_first_7taps[2][7] = 
  {{0.005412,0.069591, 0.244560, 0.360875, 0.244560, 0.069591, 0.005412},
    {-0.019479, -0.123915, -0.193555, 0.0, 0.193555, 0.123915, 0.019479}};

  double simoncelli_first_9taps[2][9] =
  {{0.000738, 0.015530, 0.090260, 0.234469, 0.318007, 0.234469, 0.090260, 0.015530, 0.000738},
    {-0.003032, -0.035241, -0.118879, -0.144383, 0.0, 0.144383, 0.118879, 0.035241, 0.003032}};

  double simoncelli_first_11taps[2][11] = 
  {{0.000097, 0.003042, 0.026178, 0.103249, 0.223725, 0.287419, 0.223725, 0.103249, 0.026178, 0.003042, 0.000097},
    {-0.000440, -0.008083, -0.044978, -0.108831, -0.112871, 0.0, 0.112871, 0.108831, 0.044978, 0.008083, 0.000440}};

  double simoncelli_second_5taps[3][5] = 
  {{0.030320, 0.249724, 0.439911, 0.249724, 0.030320},
    {-0.104550, -0.292315, 0.0, 0.292315,  0.104550},
    {0.232905, 0.002668, -0.471147, 0.002668,  0.232905}};

  double simoncelli_second_7taps[3][7] = 
  {{0.004711, 0.069321, 0.245410, 0.361117, 0.245410, 0.069321, 0.004711},
    {-0.018708, -0.125376, -0.193091, 0.0, 0.193091, 0.125376, 0.018708},
    {0.055336, 0.137778, -0.056554, -0.273118, -0.056554, 0.137778, 0.055336}};

  double simoncelli_second_9taps[3][9] = 
  {{0.000734, 0.015591, 0.090295, 0.234405, 0.317952, 0.234405, 0.090295, 0.015591, 0.000734},
    {-0.003080, -0.035311, -0.118783, -0.144393, 0.0,  0.144393, 0.118783, 0.035311, 0.003080},
    {0.010316, 0.061361, 0.085644, -0.061329, -0.191986, -0.061329, 0.085644, 0.061361, 0.010316}};

  double simoncelli_second_11taps[3][11] = 
  {{0.000097, 0.003042, 0.026178, 0.103249, 0.223725, 0.287419, 0.223725, 0.103249, 0.026178, 0.003042, 0.000097},
    {-0.000440, -0.008083, -0.044978, -0.108831, -0.112871, 0.0, 0.112871, 0.108831, 0.044978, 0.008083, 0.000440},
    {0.001688, 0.017848, 0.057382, 0.053662, -0.059065, -0.143029, -0.059065, 0.053662, 0.057382, 0.017848, 0.001688}};

  /* 
   * Gaussian with a sigma = 0.96. The Matlab code to compute the coeficients
   * for the pre-filter, first order filter and second order filter is as follows:
   *
   * sigma = 0.96; pre = 1/sqrt(2*pi*sigma^2)*exp(-x.^2/(2*sigma^2)); 
   * disp(pre); disp(x/sigma^2 .* pre); disp((x.^2/sigma^4 - 1/sigma^2) .* pre)
   *
   */

  double gauss_3taps[3][3] =
  {{0.24156, 0.41556, 0.24156},
    {-0.26211, 0.00000, 0.26211},
    {0.022297, -0.450917, 0.022297}};

  double gauss_5taps[3][5] = 
  {{0.047442, 0.241557, 0.415565, 0.241557, 0.047442},
    {-0.10295, -0.26211, 0.00000, 0.26211, 0.10295},
    {0.171949, 0.022297, -0.450917, 0.022297, 0.171949}};

  double gauss_7taps[3][7] = 
  {{0.0031482, 0.0474416, 0.2415566, 0.4155649, 0.2415566, 0.0474416, 0.0031482},
    {-0.01025, -0.10295, -0.26211, 0.00000, 0.26211, 0.10295, 0.01025},
    {0.029943, 0.171949, 0.022297, -0.450917, 0.022297, 0.171949, 0.029943}};

  double gauss_9taps[3][9] =
  {{7.0586e-05, 3.1482e-03, 4.7442e-02, 2.4156e-01, 4.1556e-01, 2.4156e-01, 4.7442e-02, 3.1482e-03, 7.0586e-05},
    {-0.00031, -0.01025, -0.10295, -0.26211, 0.00000, 0.26211, 0.10295, 0.01025, 0.00031},
    {0.0012531, 0.0299434, 0.1719490, 0.0222972, -0.4509167, 0.0222972, 0.1719490, 0.0299434, 0.0012531}};

  double gauss_11taps[3][11] = 
  {{5.3474e-07, 7.0586e-05, 3.1482e-03, 4.7442e-02, 2.4156e-01, 4.1556e-01, 2.4156e-01, 4.7442e-02, 3.1482e-03, 7.0586e-05, 5.3474e-07},
    {-0.00000, -0.00031, -0.01025, -0.10295, -0.26211, 0.00000, 0.26211, 0.10295, 0.01025, 0.00031, 0.00000},
    {1.5160e-05, 1.2531e-03, 2.9943e-02, 1.7195e-01, 2.2297e-02, -4.5092e-01, 2.2297e-02, 1.7195e-01, 2.9943e-02, 1.2531e-03, 1.5160e-05}};

  /* Code begins here */

  my_filter = NULL;

  if (filter_type == FILTER_SIMONCELLI)
  {
    if (derivative_order == D_FIRST)
    {
      if (length == 5)
	my_filter = &simoncelli_first_5taps[index][0];
      else if (length == 7)
	my_filter = &simoncelli_first_7taps[index][0];
      else if (length == 9)
	my_filter = &simoncelli_first_9taps[index][0];
      else if (length == 11)
	my_filter = &simoncelli_first_11taps[index][0];
      else 
	error("no valid filter\n");
    }
    else if (derivative_order == D_SECOND)
    {
      if (length == 5)
	my_filter = &simoncelli_second_5taps[index][0];
      else if (length == 7)
	my_filter = &simoncelli_second_7taps[index][0];
      else if (length == 9)
	my_filter = &simoncelli_second_9taps[index][0];
      else if (length == 11)
	my_filter = &simoncelli_second_11taps[index][0];
      else 
	error("no valid filter\n");
    }
    else 
      error("no valid filter\n");
  }
  else if (filter_type == FILTER_GAUSS)
  {
    if (length == 3)
      my_filter = &gauss_3taps[index][0];
    else if (length == 5)
      my_filter = &gauss_5taps[index][0];
    else if (length == 7)
      my_filter = &gauss_7taps[index][0];
    else if (length == 9)
      my_filter = &gauss_9taps[index][0];
    else if (length == 11)
      my_filter = &gauss_11taps[index][0];
    else 
      error("no valid filter\n");
  }
  else 
    error("no valid filter\n");

  /* Effective length of the filter: we add zeros to both sides
   * of the filter 
   */

  eff_length = length + 2;

  /* Memory allocation */

  filter_zero_padded = mgtn_vector_alloc(NULL, 512);
  filter_shifted = mgtn_vector_alloc(NULL, 512);

  filter = mgtn_vector_alloc(NULL, eff_length * interpolated_samples);

  /* Set of zero all samples */

  for(i = 0; i < 512; i++)
    filter_zero_padded->values[i] = 0.0;

  for(i = 0; i < 512; i++)
    filter_shifted->values[i] = 0.0;

  for(i = 0; i < filter->size; i++)
    filter->values[i] = 0.0;

  /* Initialize filterpadded */

  for(i = 1; i <= length; i++)
    filter_zero_padded->values[i] = my_filter[i - 1];

  /* Insert zero shifted filter */

  for(i = 0; i < eff_length; i++)
    filter->values[i * interpolated_samples] = filter_zero_padded->values[i];

  /* Shift filter */

  k = eff_length - 1;

  for(j = 1; j < interpolated_samples; j++)
  {
    shift = (double) j * (1.0 / (double) interpolated_samples);

    mgtn_fourier_nonintegershift(filter_zero_padded, shift, filter_shifted);

    for(i = 0; i < k; i++)
      filter->values[j + i * interpolated_samples] = filter_shifted->values[i];
  }

  /* Delete dinyamic allocated memory */

  mgtn_vector_delete(filter_zero_padded);
  mgtn_vector_delete(filter_shifted); 

  return filter;
}

/**
 * 
 *  Get filter coefficients for presmoothing and prolongation 
 *
 */

struct vector *mgtn_get_filter_presmooth(
    int filter_id, 
    int type_filter)
{
  struct filter {
    int flag;
    int length;
    double *filter;
  };

  /* Classical restriction-pronloge filter */

  double classical_rp_filter[3] = {0.25, 0.5, 0.25};
  struct filter classical_rp = {D_FILTER_RESTRICTION, 3, classical_rp_filter};

  /* Simoncelly presmoothing filter, length 5 */

  double simoncelli_5[5] = {0.030320, 0.249724, 0.439911, 0.249724, 0.030320};
  struct filter presmooth_5 = {D_FILTER_RESTRICTION, 5, simoncelli_5};

  /* Simoncelly presmoothing filter, length 7 */

  double simoncelli_7[7] = {0.005412,0.069591, 0.244560, 0.360875, 0.244560, 0.069591, 0.005412};
  struct filter presmooth_7 = {D_FILTER_RESTRICTION, 7, simoncelli_7};

  /* Simoncelly presmoothing filter, length 9 */

  double simoncelli_9[9] = {0.000738, 0.015530, 0.090260, 0.234469, 0.318007, 0.234469, 0.090260, 0.015530, 0.000738};
  struct filter presmooth_9 = {D_FILTER_RESTRICTION, 9, simoncelli_9};

  /* Simoncelly presmoothing filter, length 11 */

  double simoncelli_11[11] = {0.000097, 0.003042, 0.026178, 0.103249, 0.223725, 0.287419, 0.223725, 0.103249, 0.026178, 0.003042, 0.000097};
  struct filter presmooth_11  = {D_FILTER_RESTRICTION, 11, simoncelli_11};

  /* Interpolation filter, length 7 */

  double interpolate_7_filter[7] = {-0.06250, 0.00000, 0.56250, 1.00000, 0.56250, 0.00000, -0.06250};
  struct filter interpolate_7 = {D_FILTER_PROLONGATION, 7, interpolate_7_filter};

  /* Code begins here */

  struct filter *my_filter = NULL;

  switch (filter_id) {
    case D_CLASSICAL_RP:
      my_filter = &classical_rp;
      break;

    case D_PRESMOOTH_5:
      my_filter = &presmooth_5;
      break;

    case D_PRESMOOTH_7:
      my_filter = &presmooth_7;
      break;

    case D_PRESMOOTH_9:
      my_filter = &presmooth_9;
      break;

    case D_PRESMOOTH_11:
      my_filter = &presmooth_11;
      break;

    case D_INTERPOLATE_7:
      my_filter = &interpolate_7;
      break;

    default:
      error("filter not valid.\n");
  }

  int i;
  struct vector *filter;

  /* Allocate memory for filter */

  filter = mgtn_vector_alloc(NULL, my_filter->length);
  filter->shift = - (filter->size - 1) >> 1; 

  /* Copy desired filter */

  for(i = 0; i < my_filter->length; i++)
    filter->values[i] = my_filter->filter[i];

  /* Normalize filter taps if necessary */

  if ((type_filter == D_FILTER_PROLONGATION) && (my_filter->flag == D_FILTER_RESTRICTION))
  {
    for(i = 0; i < filter->size; i++)
      filter->values[i] *= 2.0;
  }

  if ((type_filter == D_FILTER_RESTRICTION) && (my_filter->flag == D_FILTER_PROLONGATION))
  {
    for(i = 0; i < filter->size; i++)
      filter->values[i] *= 0.5;
  }

  return filter;
}

/**
 * 
 * Compute displaced frame difference (dfd). Used for evaluation
 * purposes at the end of the optical flow algorithm. The
 * parameter border is used to exclude a border around the image
 * when computing the dfd.
 *
 */

double mgtn_dfd(int border, struct resources *DD)
{
  double term1, term2, term3, *p_u, *p_v;

  int res, offset, nrow, ncol, nb_pixels;
  int min_x_border, min_y_border, max_x_border, max_y_border, max_x, max_y;

  struct vector2 coord_pix, coord_phi;
  struct image *u, *v;

  /* Get flow for current resolution level */

  res = DD->curr_level;

  u = DD->u[res];
  v = DD->v[res];

  ncol = u->ncol;
  nrow = u->nrow;

  /* Setup borders around images */

  min_x_border = border;
  min_y_border = border;

  max_x = ncol - 1;
  max_y = nrow - 1;

  max_x_border = max_x - border;
  max_y_border = max_y - border;

  /* Initialize variables to avoid complaining of compiler */

  nb_pixels = 0; 
  term3 = 0;

  /* Loop through all the pixels */

  for(coord_pix.y = min_y_border; coord_pix.y <= max_y_border; coord_pix.y++)
  {
    offset = coord_pix.y * ncol + min_x_border;

    p_u = u->gray + offset;
    p_v = v->gray + offset;

    for(coord_pix.x = min_x_border; coord_pix.x <= max_x_border; coord_pix.x++, p_u++, p_v++)
    {
      coord_phi.y = coord_pix.y + *p_v;
      coord_phi.x = coord_pix.x + *p_u;

      if ((coord_phi.x < 0) || (coord_phi.x > max_x) || 
	  (coord_phi.y < 0) || (coord_phi.y > max_y))
	continue;

      term1 = mgtn_image_interpolate(0, &coord_pix, DD);
      term2 = mgtn_image_interpolate(1, &coord_phi, DD);

      term3 += fabs(term2 - term1);
      nb_pixels++;
    }
  }

  term3 = term3 / (double) nb_pixels;

  return term3;
}

/**
 * 
 * Compute gradient norm. This norm is not normalized by the
 * number of elements.
 *
 */

double mgtn_gradient_norm(
    struct image *u, 
    struct image *v)
{
  int i, j, ncol, nrow, offset;
  int min_x, min_y, max_x, max_y;

  double sum;
  double *p_u, *p_v;

  ncol = u->ncol;
  nrow = u->nrow;

  min_x = 1;
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  sum = 0.0;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;

    p_u = u->gray + offset;
    p_v = v->gray + offset;

    for(j = min_x; j <= max_x; j++, p_u++, p_v++)
      sum += SQR(*p_u) + SQR(*p_v);
  }

  sum = sqrt(sum);

  return sum;
}

/**
 * 
 * Check if the (ground truth) flow is unknown. Used for AAE and EPE
 * computation
 *
 */

#define UNKNOWN_FLOW_THRESH 1e9

static int cac_unknown_flow(
  double u, 
  double v) 
{
    return ((fabs(u) >  UNKNOWN_FLOW_THRESH) 
	|| (fabs(v) >  UNKNOWN_FLOW_THRESH)
	|| isnan(u) || isnan(v));
}


/**
 * 
 * Compute end point error. 
 *
 */

double mgtn_end_point_error(
    struct image *u1,
    struct image *v1, 
    struct image *u2, 
    struct image *v2)
{
  int i, j, nrow, ncol, offset, nelems;
  int min_x, min_y, max_x, max_y;

  double sum, a, b;
  double *p_u1, *p_u2, *p_v1, *p_v2;

  ncol = u1->ncol;
  nrow = u1->nrow;

  min_x = 1;
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  sum = 0.0;
  nelems = 0;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;

    p_u1 = u1->gray + offset;
    p_u2 = u2->gray + offset;

    p_v1 = v1->gray + offset;
    p_v2 = v2->gray + offset;

    for(j = min_x; j <= max_x; j++, p_u1++, p_u2++, p_v1++, p_v2++)
    {
      if ((cac_unknown_flow(*p_u1, *p_v1)) || (cac_unknown_flow(*p_u2, *p_v2)))
	continue;

      a = *p_u1 - *p_u2;
      b = *p_v1 - *p_v2;

      sum += sqrt(a*a+b*b);
      nelems++;
    }
  }

  sum /= (double) nelems;

  return sum;
}

/**
 * 
 * Compute angular error between two flows. 
 *
 */

double mgtn_average_angular_error(
    struct image *u1, 
    struct image *v1, 
    struct image *u2, 
    struct image *v2)
{
  int i, j, nrow, ncol, offset, current_size;
  int min_x, min_y, max_x, max_y;

  double esc_prod, term1, term2 ,term3, tmp, ae, mean_ae;
  double *p_u1, *p_u2, *p_v1, *p_v2;

  ncol = u1->ncol;
  nrow = u1->nrow;

  min_x = 1;
  min_y = 1;

  max_x = ncol - 2;
  max_y = nrow - 2;

  current_size = 0.0;
  mean_ae = 0.0;

  ae = 0.0;

  for(i = min_y; i <= max_y; i++)
  {
    offset = i * ncol + min_x;

    p_u1 = u1->gray + offset;
    p_u2 = u2->gray + offset;

    p_v1 = v1->gray + offset;
    p_v2 = v2->gray + offset;

    for(j = min_x; j <= max_x; j++, p_u1++, p_u2++, p_v1++, p_v2++)
    {
      if ((cac_unknown_flow(*p_u1, *p_v1)) || (cac_unknown_flow(*p_u2, *p_v2)))
	continue;

      esc_prod = *p_u1 * *p_u2 + *p_v1 * *p_v2 + 1.0;

      term1 = (*p_u1 * *p_u1 + *p_v1 * *p_v1 + 1.0);
      term2 = (*p_u2 * *p_u2 + *p_v2 * *p_v2 + 1.0);
      term3 = sqrt(term1 * term2);

      tmp = esc_prod / term3;

      tmp = MIN(tmp, 1.0);
      tmp = MAX(tmp, -1.0);

      ae = acos(tmp);

      current_size++;

      mean_ae  = mean_ae  + ((double)ae - mean_ae) / ((double)current_size);
    }
  }

  /* Transform angular errot to degrees */

  mean_ae =  (mean_ae * 360.0) / (2.0 * M_PI);

  return mean_ae;
}


